"""Shared constants for IPv4 and IPv6."""

# Protocol numbers - http://www.iana.org/assignments/protocol-numbers
IP_PROTO_IP = 0  # dummy for IP
IP_PROTO_HOPOPTS = IP_PROTO_IP  # IPv6 hop-by-hop options
IP_PROTO_ICMP = 1  # ICMP
IP_PROTO_IGMP = 2  # IGMP
IP_PROTO_GGP = 3  # gateway-gateway protocol
IP_PROTO_IPIP = 4  # IP in IP
IP_PROTO_ST = 5  # ST datagram mode
IP_PROTO_TCP = 6  # TCP
IP_PROTO_CBT = 7  # CBT
IP_PROTO_EGP = 8  # exterior gateway protocol
IP_PROTO_IGP = 9  # interior gateway protocol
IP_PROTO_BBNRCC = 10  # BBN RCC monitoring
IP_PROTO_NVP = 11  # Network Voice Protocol
IP_PROTO_PUP = 12  # PARC universal packet
IP_PROTO_ARGUS = 13  # ARGUS
IP_PROTO_EMCON = 14  # EMCON
IP_PROTO_XNET = 15  # Cross Net Debugger
IP_PROTO_CHAOS = 16  # Chaos
IP_PROTO_UDP = 17  # UDP
IP_PROTO_MUX = 18  # multiplexing
IP_PROTO_DCNMEAS = 19  # DCN measurement
IP_PROTO_HMP = 20  # Host Monitoring Protocol
IP_PROTO_PRM = 21  # Packet Radio Measurement
IP_PROTO_IDP = 22  # Xerox NS IDP
IP_PROTO_TRUNK1 = 23  # Trunk-1
IP_PROTO_TRUNK2 = 24  # Trunk-2
IP_PROTO_LEAF1 = 25  # Leaf-1
IP_PROTO_LEAF2 = 26  # Leaf-2
IP_PROTO_RDP = 27  # "Reliable Datagram" proto
IP_PROTO_IRTP = 28  # Inet Reliable Transaction
IP_PROTO_TP = 29  # ISO TP class 4
IP_PROTO_NETBLT = 30  # Bulk Data Transfer
IP_PROTO_MFPNSP = 31  # MFE Network Services
IP_PROTO_MERITINP = 32  # Merit Internodal Protocol
IP_PROTO_SEP = 33  # Sequential Exchange proto
IP_PROTO_3PC = 34  # Third Party Connect proto
IP_PROTO_IDPR = 35  # Interdomain Policy Route
IP_PROTO_XTP = 36  # Xpress Transfer Protocol
IP_PROTO_DDP = 37  # Datagram Delivery Proto
IP_PROTO_CMTP = 38  # IDPR Ctrl Message Trans
IP_PROTO_TPPP = 39  # TP++ Transport Protocol
IP_PROTO_IL = 40  # IL Transport Protocol
IP_PROTO_IP6 = 41  # IPv6
IP_PROTO_SDRP = 42  # Source Demand Routing
IP_PROTO_ROUTING = 43  # IPv6 routing header
IP_PROTO_FRAGMENT = 44  # IPv6 fragmentation header
IP_PROTO_RSVP = 46  # Reservation protocol
IP_PROTO_GRE = 47  # General Routing Encap
IP_PROTO_MHRP = 48  # Mobile Host Routing
IP_PROTO_ENA = 49  # ENA
IP_PROTO_ESP = 50  # Encap Security Payload
IP_PROTO_AH = 51  # Authentication Header
IP_PROTO_INLSP = 52  # Integated Net Layer Sec
IP_PROTO_SWIPE = 53  # SWIPE
IP_PROTO_NARP = 54  # NBMA Address Resolution
IP_PROTO_MOBILE = 55  # Mobile IP, RFC 2004
IP_PROTO_TLSP = 56  # Transport Layer Security
IP_PROTO_SKIP = 57  # SKIP
IP_PROTO_ICMP6 = 58  # ICMP for IPv6
IP_PROTO_NONE = 59  # IPv6 no next header
IP_PROTO_DSTOPTS = 60  # IPv6 destination Woptions
IP_PROTO_ANYHOST = 61  # any host internal proto
IP_PROTO_CFTP = 62  # CFTP
IP_PROTO_ANYNET = 63  # any local network
IP_PROTO_EXPAK = 64  # SATNET and Backroom EXPAK
IP_PROTO_KRYPTOLAN = 65  # Kryptolan
IP_PROTO_RVD = 66  # MIT Remote Virtual Disk
IP_PROTO_IPPC = 67  # Inet Pluribus Packet Core
IP_PROTO_DISTFS = 68  # any distributed fs
IP_PROTO_SATMON = 69  # SATNET Monitoring
IP_PROTO_VISA = 70  # VISA Protocol
IP_PROTO_IPCV = 71  # Inet Packet Core Utility
IP_PROTO_CPNX = 72  # Comp Proto Net Executive
IP_PROTO_CPHB = 73  # Comp Protocol Heart Beat
IP_PROTO_WSN = 74  # Wang Span Network
IP_PROTO_PVP = 75  # Packet Video Protocol
IP_PROTO_BRSATMON = 76  # Backroom SATNET Monitor
IP_PROTO_SUNND = 77  # SUN ND Protocol
IP_PROTO_WBMON = 78  # WIDEBAND Monitoring
IP_PROTO_WBEXPAK = 79  # WIDEBAND EXPAK
IP_PROTO_EON = 80  # ISO CNLP
IP_PROTO_VMTP = 81  # Versatile Msg Transport
IP_PROTO_SVMTP = 82  # Secure VMTP
IP_PROTO_VINES = 83  # VINES
IP_PROTO_TTP = 84  # TTP
IP_PROTO_NSFIGP = 85  # NSFNET-IGP
IP_PROTO_DGP = 86  # Dissimilar Gateway Proto
IP_PROTO_TCF = 87  # TCF
IP_PROTO_EIGRP = 88  # EIGRP
IP_PROTO_OSPF = 89  # Open Shortest Path First
IP_PROTO_SPRITERPC = 90  # Sprite RPC Protocol
IP_PROTO_LARP = 91  # Locus Address Resolution
IP_PROTO_MTP = 92  # Multicast Transport Proto
IP_PROTO_AX25 = 93  # AX.25 Frames
IP_PROTO_IPIPENCAP = 94  # yet-another IP encap
IP_PROTO_MICP = 95  # Mobile Internet Ctrl
IP_PROTO_SCCSP = 96  # Semaphore Comm Sec Proto
IP_PROTO_ETHERIP = 97  # Ethernet in IPv4
IP_PROTO_ENCAP = 98  # encapsulation header
IP_PROTO_ANYENC = 99  # private encryption scheme
IP_PROTO_GMTP = 100  # GMTP
IP_PROTO_IFMP = 101  # Ipsilon Flow Mgmt Proto
IP_PROTO_PNNI = 102  # PNNI over IP
IP_PROTO_PIM = 103  # Protocol Indep Multicast
IP_PROTO_ARIS = 104  # ARIS
IP_PROTO_SCPS = 105  # SCPS
IP_PROTO_QNX = 106  # QNX
IP_PROTO_AN = 107  # Active Networks
IP_PROTO_IPCOMP = 108  # IP Payload Compression
IP_PROTO_SNP = 109  # Sitara Networks Protocol
IP_PROTO_COMPAQPEER = 110  # Compaq Peer Protocol
IP_PROTO_IPXIP = 111  # IPX in IP
IP_PROTO_VRRP = 112  # Virtual Router Redundancy
IP_PROTO_PGM = 113  # PGM Reliable Transport
IP_PROTO_ANY0HOP = 114  # 0-hop protocol
IP_PROTO_L2TP = 115  # Layer 2 Tunneling Proto
IP_PROTO_DDX = 116  # D-II Data Exchange (DDX)
IP_PROTO_IATP = 117  # Interactive Agent Xfer
IP_PROTO_STP = 118  # Schedule Transfer Proto
IP_PROTO_SRP = 119  # SpectraLink Radio Proto
IP_PROTO_UTI = 120  # UTI
IP_PROTO_SMP = 121  # Simple Message Protocol
IP_PROTO_SM = 122  # SM
IP_PROTO_PTP = 123  # Performance Transparency
IP_PROTO_ISIS = 124  # ISIS over IPv4
IP_PROTO_FIRE = 125  # FIRE
IP_PROTO_CRTP = 126  # Combat Radio Transport
IP_PROTO_CRUDP = 127  # Combat Radio UDP
IP_PROTO_SSCOPMCE = 128  # SSCOPMCE
IP_PROTO_IPLT = 129  # IPLT
IP_PROTO_SPS = 130  # Secure Packet Shield
IP_PROTO_PIPE = 131  # Private IP Encap in IP
IP_PROTO_SCTP = 132  # Stream Ctrl Transmission
IP_PROTO_FC = 133  # Fibre Channel
IP_PROTO_RSVPIGN = 134  # RSVP-E2E-IGNORE
IP_PROTO_RAW = 255  # Raw IP packets
IP_PROTO_RESERVED = IP_PROTO_RAW  # Reserved
IP_PROTO_MAX = 255



"""
Internet Control Message Protocol for IPv4.
https://www.iana.org/assignments/icmp-parameters/icmp-parameters.xhtml
https://tools.ietf.org/html/rfc792
"""
import logging

from packetracer import packetracer, checksum
from packetracer.packetracer_meta import FIELD_FLAG_AUTOUPDATE, FIELD_FLAG_IS_TYPEFIELD

logger = logging.getLogger("packetracer")

# Types (icmp_type) and codes (icmp_code) -
# http://www.iana.org/assignments/icmp-parameters


ICMP_ECHO_REPLY = 0  # echo reply
ICMP_UNREACH = 3  # dest unreachable
ICMP_SRCQUENCH = 4  # packet lost, slow down
ICMP_REDIRECT = 5  # shorter route
ICMP_ALTHOSTADDR = 6  # alternate host address
ICMP_ECHO = 8  # echo service
ICMP_RTRADVERT = 9  # router advertise
ICMP_RTRSEL = 10  # router selection
ICMP_TIMEXCEED = 11  # time exceeded, code:
ICMP_PARAMPROB = 12  # ip header bad
ICMP_TSTAMP = 13  # timestamp request
ICMP_TSTAMPREPLY = 14  # timestamp reply
ICMP_INFO = 15  # information request
ICMP_INFOREPLY = 16  # information reply
ICMP_MASK = 17  # address mask request
ICMP_MASKREPLY = 18  # address mask reply
ICMP_TRACEROUTE = 30  # traceroute
ICMP_DATACONVERR = 31  # data conversion error
ICMP_MOBILE_REDIRECT = 32  # mobile host redirect
ICMP_IP6_WHEREAREYOU = 33  # IPv6 where-are-you
ICMP_IP6_IAMHERE = 34  # IPv6 i-am-here
ICMP_MOBILE_REG = 35  # mobile registration req
ICMP_MOBILE_REGREPLY = 36  # mobile registration reply
ICMP_DNS = 37  # domain name request
ICMP_DNSREPLY = 38  # domain name reply
ICMP_SKIP = 39  # SKIP
ICMP_PHOTURIS = 40  # Photuris


class ICMP(packetracer.Packet):
    __hdr__ = (
        ("type", "B", ICMP_ECHO, FIELD_FLAG_IS_TYPEFIELD),
        ("code", "B", 0),
        ("sum", "H", 0, FIELD_FLAG_AUTOUPDATE)
    )
    type_t = packetracer.get_property_translator("type", "ICMP_")

    def _update_fields(self):
        # logger.debug("sum is: %d" % self.sum)
        if self.sum_au_active and self._changed():
            # logger.debug("sum is: %d" % self.sum)
            # logger.debug("header: %r", self.header_bytes)
            # logger.debug("body: %r", self.body_bytes)
            self.sum = 0
            self.sum = checksum.in_cksum(self.header_bytes + self.body_bytes)
        # logger.debug("sum is: %d" % self.sum)

    def _dissect(self, buf):
        # logger.debug("ICMP: adding fields for type: %d" % buf[0])
        self._init_handler(buf[0], buf[4:])
        return 4

    class Echo(packetracer.Packet):
        __hdr__ = (
            ("id", "H", 0),
            ("seq", "H", 1),
            ("ts", "Q", 0)
        )

        def _dissect(self, buf):
            hlen = 12

            if len(buf) < 12:
                # not enough bytes for ts
                self.ts = None
                hlen = 4
            return hlen

    class Unreach(packetracer.Packet):
        __hdr__ = (
            ("pad", "I", 0),
        )

        CODE_UNREACH_NET = 0  # bad net
        CODE_UNREACH_HOST = 1  # bad host
        CODE_UNREACH_PROTO = 2  # bad protocol
        CODE_UNREACH_PORT = 3  # bad port
        CODE_UNREACH_NEEDFRAG = 4  # IP_DF caused drop
        CODE_UNREACH_SRCFAIL = 5  # src route failed
        CODE_UNREACH_NET_UNKNOWN = 6  # unknown net
        CODE_UNREACH_HOST_UNKNOWN = 7  # unknown host
        CODE_UNREACH_ISOLATED = 8  # src host isolated
        CODE_UNREACH_NET_PROHIB = 9  # for crypto devs
        CODE_UNREACH_HOST_PROHIB = 10  # ditto
        CODE_UNREACH_TOSNET = 11  # bad tos for net
        CODE_UNREACH_TOSHOST = 12  # bad tos for host
        CODE_UNREACH_FILTER_PROHIB = 13  # prohibited access
        CODE_UNREACH_HOST_PRECEDENCE = 14  # precedence error
        CODE_UNREACH_PRECEDENCE_CUTOFF = 15  # precedence cutoff

    class Quench(packetracer.Packet):
        __hdr__ = (
            ("pad", "I", 0),
        )

    class Redirect(packetracer.Packet):
        __hdr__ = (
            ("gw", "I", 0),
        )
        CODE_REDIRECT_NET = 0  # for network
        CODE_REDIRECT_HOST = 1  # for host
        CODE_REDIRECT_TOSNET = 2  # for tos and net
        CODE_REDIRECT_TOSHOST = 3  # for tos and host

    class RouterAdvertisement(packetracer.Packet):
        __hdr__ = (
            ("numaddr", "B", 0),
            ("addrsize", "B", 0),
            ("lifetime", "H", 0)
        )

        CODE_RTRADVERT_NORMAL = 0  # normal
        CODE_RTRADVERT_NOROUTE_COMMON = 16  # selective routing
        CODE_RTRSOLICIT = 10  # router solicitation

    class RouterSelection(packetracer.Packet):
        __hdr__ = (
            ("numaddr", "B", 0),
            ("addrsize", "B", 0),
            ("lifetime", "H", 0)
        )

    class TimeExceed(packetracer.Packet):
        __hdr__ = (
            ("pad", "I", 0),
        )

        CODE_TIMEXCEED_INTRANS = 0  # ttl==0 in transit
        CODE_TIMEXCEED_REASS = 1  # ttl==0 in reass

    class ParamProblem(packetracer.Packet):
        __hdr__ = (
            ("pointer", "B", 0),
            ("unused", "3s", b"\x00" * 3)
        )

        CODE_PARAMPROB_ERRATPTR = 0  # req. opt. absent
        CODE_PARAMPROB_OPTABSENT = 1  # req. opt. absent
        CODE_PARAMPROB_LENGTH = 2  # bad length

    class Photuris(packetracer.Packet):
        class ParamProblem(packetracer.Packet):
            __hdr__ = (
                ("reserved", "H", 0),
                ("pointer", "H", 0)
            )

        CODE_PHOTURIS_UNKNOWN_INDEX = 0  # unknown sec index
        CODE_PHOTURIS_AUTH_FAILED = 1  # auth failed
        CODE_PHOTURIS_DECOMPRESS_FAILED = 2  # decompress failed
        CODE_PHOTURIS_DECRYPT_FAILED = 3  # decrypt failed
        CODE_PHOTURIS_NEED_AUTHN = 4  # no authentication
        CODE_PHOTURIS_NEED_AUTHZ = 5  # no authorization

    __handler__ = {
        (ICMP_ECHO, ICMP_ECHO_REPLY): Echo,
        ICMP_UNREACH: Unreach,
        ICMP_SRCQUENCH: Quench,
        ICMP_REDIRECT: Redirect,
        ICMP_RTRADVERT: RouterAdvertisement,
        ICMP_RTRSEL: RouterSelection,
        ICMP_TIMEXCEED: TimeExceed,
        ICMP_PARAMPROB: ParamProblem,
        ICMP_PHOTURIS: Photuris
    }



"""
Simple Traversal of UDP through NAT (STUN).
RFC 3489
http://tools.ietf.org/html/rfc3489
"""
import logging

from packetracer import triggerlist
from packetracer.packetracer import Packet
from packetracer.structcbs import unpack_H

logger = logging.getLogger("packetracer")

# Message Types
BINDING_REQUEST = 0x0001
BINDING_RESPONSE = 0x0101
BINDING_ERROR_RESPONSE = 0x0111
SHARED_SECRET_REQUEST = 0x0002
SHARED_SECRET_RESPONSE = 0x0102
SHARED_SECRET_ERROR_RESPONSE = 0x0112

# Message Attributes
MAPPED_ADDRESS = 0x0001
RESPONSE_ADDRESS = 0x0002
CHANGE_REQUEST = 0x0003
SOURCE_ADDRESS = 0x0004
CHANGED_ADDRESS = 0x0005
USERNAME = 0x0006
PASSWORD = 0x0007
MESSAGE_INTEGRITY = 0x0008
ERROR_CODE = 0x0009
UNKNOWN_ATTRIBUTES = 0x000A
REFLECTED_FROM = 0x000B


class StunAttr(Packet):
    __hdr__ = (
        ("type", "H", 0),
        ("len", "H", 0),
    )


class STUN(Packet):
    # 20 byte header followed by 0 or more attribute TLVs.
    __hdr__ = (
        ("type", "H", 0),
        ("len", "H", 0),
        ("cookie", "I", 0),
        ("xid", "12s", b"\x00" * 14),
        ("attrs", None, triggerlist.TriggerList)
    )

    @staticmethod
    def __parse_attrs(buf):
        attributes = []
        off = 0

        # t:2 l:2 v:x
        while off < len(buf):
            l_content = unpack_H(buf[off + 2: off + 4])[0]
            padding = (4 - (l_content % 4)) % 4
            l_total = l_content + padding + 2 + 2
            # logger.debug("STUN attr l_content: %d, padding: %d, value: %s" %
            #	 (l_content, padding, buf[off : off + l_total]))
            attributes.append(StunAttr(buf[off: off + l_total]))
            off += l_total
        return attributes

    def _dissect(self, buf):
        # logger.debug("dissecting: %s" % buf)
        self._init_triggerlist("attrs", buf[20:], self.__parse_attrs)
        return len(buf)



"""
Utility functions.
"""
import ipaddress
import logging
import math
import os
import re
import subprocess

from packetracer import packetracer as packetracer

logger = logging.getLogger("packetracer")

try:
    import netifaces
except ImportError:
    logger.warning("Couldn't load netifaces, some utils won't work")

log = math.log
mac_bytes_to_str = packetracer.mac_bytes_to_str


def switch_wlan_channel(iface, channel, shutdown_prior=False):
    """
    Switch wlan channel to channel.
    Requirements: ifconfig, iwconfig

    iface -- interface name
    channel -- channel numer to be set as number
    shutdown_prior -- shut down interface prior to setting channel
    """
    if shutdown_prior:
        cmd_call = ["ifconfig", iface, "down"]
        subprocess.check_call(cmd_call)

    cmd_call = ["iwconfig", iface, "channel", "%d" % channel]
    subprocess.check_call(cmd_call)

    if shutdown_prior:
        cmd_call = ["ifconfig", iface, "up"]
        subprocess.check_call(cmd_call)


WLAN_MODE_MANAGED = 0
WLAN_MODE_MONITOR = 1
WLAN_MODE_UNKNOWN = 2

_MODE_STR_INT_TRANSLATE = {
    b"managed": WLAN_MODE_MANAGED,
    b"monitor": WLAN_MODE_MONITOR,
    b"": WLAN_MODE_UNKNOWN
}

PATTERN_MODE = re.compile(br"Mode:(\w+) ")


def get_wlan_mode(iface):
    """
    return -- [MODE_MANAGED | MODE_MONITOR | MODE_UNKNOWN]
    """
    cmd_call = ["iwconfig", iface]
    output = subprocess.check_output(cmd_call)
    match = PATTERN_MODE.search(output)

    found_str = match.group(1).lower()
    return _MODE_STR_INT_TRANSLATE.get(found_str, WLAN_MODE_UNKNOWN)


def is_interface_up(iface):
    """
    return -- [True | False]
    """
    cmd_call = ["ifconfig"]
    pattern_up = re.compile(b"^" + bytes(iface, "UTF-8") + b": flags=", re.MULTILINE)
    output = subprocess.check_output(cmd_call)
    return pattern_up.search(output) is not None


def set_interface_mode(iface, monitor_active=None, mtu=None, state_active=None):
    """
    Configure an interface, primarily for wifi monitor mode
    Requirements: ifconfig, iwconfig

    monitor_active -- activate/deactivate monitor mode (only for wlan interfaces)
    state_active -- set interface state
    """
    initial_state_up = is_interface_up(iface)

    if monitor_active is not None:
        cmd_call = ["ifconfig", iface, "down"]
        subprocess.check_call(cmd_call)
        mode = "monitor" if monitor_active else "managed"
        cmd_call = ["iwconfig", iface, "mode", mode]
        subprocess.check_call(cmd_call)

    if type(mtu) is int:
        cmd_call = ["ifconfig", iface, "mtu", "%d" % mtu]
        subprocess.check_call(cmd_call)

    # try:
    #	cmd_call = ["iwconfig", iface, "retry", "0"]
    #	subprocess.check_call(cmd_call)
    #	# we don't need retry but this can improve performance
    # except:
    #	# not implemented: don't care
    #	pass

    if state_active or initial_state_up:
        cmd_call = ["ifconfig", iface, "up"]
        subprocess.check_call(cmd_call)


def is_interface_present(iface_name):
    try:
        netifaces.ifaddresses(iface_name)
        return True
    except ValueError:
        # raised if interface is not present
        return False


def set_interface_state(iface_name, state_active=True):
    state_str = "up" if state_active else "down"
    output = subprocess.getoutput("ip link set dev %s %s" % (iface_name, state_str))
    logger.info(output)


PROG_CHANNEL = re.compile(br"Channel ([\d]+) :")


def get_available_wlan_channels(iface):
    """
    Requirements: iwlist

    return -- channels as integer list
    """
    cmd_call = ["iwlist", iface, "channel"]
    output = subprocess.check_output(cmd_call)
    # logger.debug("iwlist output: %r", output)

    return [int(ch) for ch in PROG_CHANNEL.findall(output)]


def set_ethernet_address(iface, ethernet_addr):
    """
    iface -- interface name
    ethernet_addr -- Ethernet address like "AA:BB:CC:DD:EE:FF"
    """
    initial_state_up = is_interface_up(iface)
    cmd_call = ["ifconfig", iface, "down"]
    subprocess.check_call(cmd_call)
    cmd_call = ["ifconfig", iface, "hw", "ether", ethernet_addr]
    subprocess.check_call(cmd_call)

    if initial_state_up:
        cmd_call = ["ifconfig", iface, "up"]
        subprocess.check_call(cmd_call)


MAC_VENDOR = {}
PROG_MACVENDOR = re.compile(r"([\w\-]{8,8})   \(hex\)\t\t(.+)")
PROG_MACVENDOR_STRIPPED = re.compile(r"(.{6,6}) (.+)")

current_dir = os.path.dirname(os.path.realpath(__file__)) + "/"

FILE_OUI = current_dir + "oui.txt"
FILE_OUI_STRIPPED = current_dir + "oui_stripped.txt"


def _convert():
    """
    Convert oui file
    return -- True on success, False otherwise
    """
    # logger.debug("loading oui file %s", FILE_OUI)

    try:
        with open(FILE_OUI, "r") as fh_read:
            for line in fh_read:
                hex_vendor = PROG_MACVENDOR.findall(line)

                if len(hex_vendor) > 0:
                    # print(hex_vendor)
                    MAC_VENDOR[hex_vendor[0][0].replace("-", "")] = hex_vendor[0][1]
    except:
        # logger.debug("no oui file present -> nothing to convert")
        return False

    try:
        with open(FILE_OUI_STRIPPED, "w") as fh_write:
            for mac, descr in MAC_VENDOR.items():
                fh_write.write("%s %s\n" % (mac, descr))
    except Exception as ex:
        logger.warning("could not create stripped oui file %r", ex)
        return False
    return True


def _load_mac_vendor():
    """
    Load oui.txt containing mac->vendor mappings into MAC_VENDOR dictionary.
    See http://standards.ieee.org/develop/regauth/oui/oui.txt
    """
    if not os.path.isfile(FILE_OUI_STRIPPED):
        success = False

        if os.path.isfile(FILE_OUI):
            success = _convert()

        if not success:
            return

    # logger.debug("loading stripped oui file %s", FILE_OUI_STRIPPED)

    try:
        with open(FILE_OUI_STRIPPED, "r") as fh_read:
            for line in fh_read:
                hex_vendor = PROG_MACVENDOR_STRIPPED.findall(line)

                if len(hex_vendor) > 0:
                    # print(hex_vendor)
                    MAC_VENDOR[hex_vendor[0][0]] = hex_vendor[0][1]
    # logger.debug("got %d vendor entries", len(MAC_VENDOR))
    except Exception as ex:
        logger.warning("could not load stripped oui file %r", ex)


def get_vendor_for_mac(mac):
    """
    mac -- First three bytes of mac address at minimum eg "AA:BB:CC...", "AABBCC..." or
        byte representation b"\xaa\xbb\xcc\xdd\xee\xff"
    return -- found vendor string or empty string
    """
    if len(MAC_VENDOR) == 1:
        return ""

    if len(MAC_VENDOR) == 0:
        _load_mac_vendor()
        # avoid loading next time
        if len(MAC_VENDOR) == 0:
            MAC_VENDOR["test"] = "test"

    if type(mac) == bytes:
        # b"\xaa\xbb\xcc\xdd\xee\xff" -> AA:BB:CC:DD:EE:FF -> AABBCC"
        mac = packetracer.mac_bytes_to_str(mac)[0:8].replace(":", "")
    else:
        # AA:BB:CC -> AABBCC
        mac = str.upper(mac.replace(":", "")[0:6])

    # logger.debug("searching mac %s", mac)
    return MAC_VENDOR.get(mac, "")


def is_special_mac(mac_str):
    """
    Check if this is a special MAC adress (not a client address). Every MAC not found
    in the official OUI database is assumed to be non-client.

    mac_str -- Uppercase mac string like "AA:BB:CC[:DD:EE:FF]", first 3 MAC-bytes are enough
    """
    return len(get_vendor_for_mac(mac_str)) == 0


ENTROPY_GRANULARITY_QUADRUPLE = 0


def get_entropy(bts, granularity):
    """
    Calcualte entropy of bts

    granularity -- ENTROPY_GRANULARITY_QUADRUPLE
    return -- entropy
    """
    symbol_count = {}
    symbol_len = 0
    if granularity == ENTROPY_GRANULARITY_QUADRUPLE:
        symbol_amount = 16

        for bt in bts:
            q1 = bt >> 4
            q2 = bt & 0x0F

            for val in [q1, q2]:
                try:
                    symbol_count[val] += 1
                except:
                    symbol_count[val] = 1

        symbol_len = len(bts) * 2  # 2 quadruples per byte
    else:
        logger.warning("invalid granularity: %d", granularity)
        return -1

    entropy = 0
    # symbol_amount = len(symbol_count)

    for _, count in symbol_count.items():
        p = count / symbol_len
        entropy += -log(p, symbol_amount) * p

    return entropy


def get_mac_for_iface(iface_name):
    """
    return -- MAC address of the interface iface_name
    Assume MAC address is always retrievable
    """
    try:
        return netifaces.ifaddresses(iface_name)[netifaces.AF_LINK][0]["addr"]
    except:
        return None


def get_ipv4_for_iface(iface_name, idx=0):
    """
    return -- IPv4 address found for interface iface_name at index idx
    """
    try:
        return netifaces.ifaddresses(iface_name)[netifaces.AF_INET][idx]["addr"]
    except:
        return None


def get_ipv4_addressinfo(iface_name, idx=0):
    """
    iface_name -- Name of the interface to get the information from
    idx -- Index to the n'th element in the address-info list (useful if multiple IP addresses are assigned)
    return -- Adressinfo (IP address, mask, broadcast address) for the given interface name
        like ("1.2.3.4", "255.255.255.0", "192.168.0.255") or None on error
    """
    try:
        addressinfo = netifaces.ifaddresses(iface_name)[netifaces.AF_INET][idx]
        return addressinfo["addr"], addressinfo["netmask"], addressinfo["broadcast"]
    except:
        return None


def nwmask_to_cidr(nmask):
    """
    TODO: Detect if IPv4 or IPv6
    nmask -- An IPv4 network mask like "255.255.255.0"
    return -- The amount of network bits in CIDR format like 24
    """
    return ipaddress.IPv4Network("1.2.3.4/%s" % nmask, strict=False).prefixlen


def get_ipv6_for_iface(iface_name, idx=0):
    """
    return -- IPv6 address found for interface iface_name at index idx
    """
    try:
        return netifaces.ifaddresses(iface_name)[netifaces.AF_INET6][idx]["addr"]
    except:
        return None


def get_gwip_for_iface(iface_name):
    """
    return -- IPv4 address of the default gateway like "1.2.3.4" for interface iface_name or None
    """
    gws = netifaces.gateways()
    gws_ipv4 = gws.get(netifaces.AF_INET, None)

    if gws_ipv4 is None:
        return None
    gw_ipv4 = None

    for gw_info in gws_ipv4:
        if iface_name in gw_info:
            gw_ipv4 = gw_info[0]
            break
    return gw_ipv4


def get_arp_cache_entry(ipaddr):
    """
    return -- MAC address for IP addess like "1.2.3.4"
    """
    mac = None
    pattern_mac = re.compile("([0-9A-Fa-f]{2}[:]){5}([0-9A-Fa-f]{2})")

    with open("/proc/net/arp", "r") as fd:
        for line in fd:
            if line.startswith(ipaddr + " "):
                mac = pattern_mac.search(line).group(0)
                break
    return mac


def add_arp_entry(ip_address, mac_address, interface_name):
    """
    Add an arp entry using linux "arp" command.
    """
    cmd_call = ["arp", "-s", ip_address, "-i", interface_name, mac_address]
    subprocess.check_call(cmd_call)


def flush_arp_cache():
    """
    Remove all arp entries from cache using linux "ip" command.
    """
    cmd_call = ["ip", "-s", "neigh", "flush", "all"]
    subprocess.check_call(cmd_call)



"""Open Shortest Path First."""

from packetracer import packetracer, checksum
from packetracer.packetracer_meta import FIELD_FLAG_AUTOUPDATE

AUTH_NONE = 0
AUTH_PASSWORD = 1
AUTH_CRYPTO = 2


class OSPF(packetracer.Packet):
    __hdr__ = (
        ("v", "B", 0),
        ("type", "B", 0),
        ("len", "H", 0),
        ("router", "I", 0),
        ("area", "I", 0),
        ("sum", "H", 0, FIELD_FLAG_AUTOUPDATE),  # _sum = sum
        ("atype", "H", 0),
        ("auth", "8s", b"")
    )

    def _update_fields(self):
        if self.sum_au_active and self._changed():
            self.sum = 0
            self.sum = checksum.in_cksum(packetracer.Packet.bin(self))



"""
Stream Control Transmission Protocol.
http://tools.ietf.org/html/rfc3286
http://tools.ietf.org/html/rfc2960
"""
import logging

from packetracer import packetracer, triggerlist, checksum
# handler
from packetracer.layer567 import diameter
from packetracer.packetracer_meta import FIELD_FLAG_AUTOUPDATE
from packetracer.structcbs import unpack_H, unpack_I

logger = logging.getLogger("packetracer")

# Chunk Types
DATA = 0
INIT = 1
INIT_ACK = 2
SACK = 3
HEARTBEAT = 4
HEARTBEAT_ACK = 5
ABORT = 6
SHUTDOWN = 7
SHUTDOWN_ACK = 8
ERROR = 9
COOKIE_ECHO = 10
COOKIE_ACK = 11
ECNE = 12
CWR = 13
SHUTDOWN_COMPLETE = 14


class Chunk(packetracer.Packet):
    __hdr__ = (
        ("type", "B", INIT),
        ("flags", "B", 0),
        ("len", "H", 0)  # length of header + data = 4 + x Bytes
    )


class SCTP(packetracer.Packet):
    __hdr__ = (
        ("sport", "H", 0),
        ("dport", "H", 0),
        ("vtag", "I", 0),
        ("sum", "I", 0, FIELD_FLAG_AUTOUPDATE),
        ("chunks", None, triggerlist.TriggerList)
    )

    # handle padding attribute
    def __get_padding(self):
        try:
            return self._padding
        except:
            return b""

    def __set_padding(self, padding):
        self._padding = padding

    padding = property(__get_padding, __set_padding)

    __handler__ = {
        123: diameter.Diameter,
    }

    def _dissect(self, buf):
        # parse chunks
        chunks = []
        off = 12
        blen = len(buf)

        # logger.debug("SCTP: parsing chunks")
        chunktype = -1

        # TODO: use lazy dissect, possible?
        while off + 4 < blen:
            dlen = unpack_H(buf[off + 2: off + 4])[0]
            # check for padding (this should be a data chunk)
            if off + dlen < blen:
                self.padding = buf[off + dlen:]
            # logger.debug("found padding: %s" % self.padding)

            chunk = Chunk(buf[off: off + dlen])
            # logger.debug("SCTP: Chunk; %s " % chunk)
            chunks.append(chunk)

            # get payload chunktype from DATA chunks
            if chunk.type == 0:
                chunktype = unpack_I(buf[off + chunk.header_len + 8: off + chunk.header_len + 8 + 4])[0]
                # logger.debug("got DATA chunk, chunktype: %d" % chunktype)
                # remove data from chunk: use bytes for handler
                chunk.body_bytes = b""
                off += len(chunk)
                # assume DATA is the last chunk
                break

            off += dlen

        self.chunks.extend(chunks)

        chunktype = unpack_H(buf[2: 4])[0]
        self._init_handler(chunktype, buf[off:-len(self.padding)])
        return off

    def _update_fields(self):
        if self.sum_au_active and self._changed():
            # logger.debug("updating checksum")
            self._calc_sum()

    def bin(self, update_auto_fields=True):
        return packetracer.Packet.bin(self, update_auto_fields=update_auto_fields) + self.padding

    def _calc_sum(self):
        # mark as changed
        self.sum = 0
        s = checksum.crc32_add(0xFFFFFFFF, self._pack_header())
        padlen = len(self.padding)

        if padlen == 0:
            s = checksum.crc32_add(s, self.body_bytes)
        else:
            # logger.debug("checksum with padding")
            s = checksum.crc32_add(s, self.body_bytes[:-padlen])

        self.sum = checksum.crc32_done(s)

    def direction(self, other):
        # logger.debug("checking direction: %s<->%s" % (self, other))
        if self.sport == other.sport and self.dport == other.dport:
            # consider packet to itself: can be DIR_REV
            return packetracer.Packet.DIR_SAME | packetracer.Packet.DIR_REV
        if self.sport == other.dport and self.dport == other.sport:
            return packetracer.Packet.DIR_REV
        return packetracer.Packet.DIR_UNKNOWN

    def reverse_address(self):
        self.sport, self.dport = self.dport, self.sport



"""
Hypertext Transfer Protocol.
"""
import logging
import re

from packetracer import packetracer, triggerlist

logger = logging.getLogger("packetracer")


class HTTPHeader(triggerlist.TriggerList):
    def _pack(self, tuple_entry):
        # logger.debug("packing HTTP-header")
        # no header = no CRNL
        if len(self) == 0:
            # logger.debug("empty buf 2")
            return b""
        # return b"\r\n".join([b": ".join(keyval) for keyval in self]) + b"\r\n\r\n"
        # logger.debug("adding: %r" % (tuple_entry[0] +b": "+ tuple_entry[1] + b"\r\n"))
        return tuple_entry[0] + b": " + tuple_entry[1] + b"\r\n"


# [Method] [Path] HTTP...\r\n
# key: value\r\n
# \r\n
# [body]
PROG_STARTLINE = re.compile(rb"[\w\./]{3,10} +[\w\./]{1,400} +[\w\./]{1,20}.+")
PROG_SPLIT_HEADBODY = re.compile(b"\r\n\r\n")
split_headbody = PROG_SPLIT_HEADBODY.split
PROG_SPLIT_HEADER = re.compile(b"\r\n")
split_header = PROG_SPLIT_HEADER.split
PROG_SPLIT_KEYVAL = re.compile(b": ")
split_keyval = PROG_SPLIT_KEYVAL.split


class HTTP(packetracer.Packet):
    __hdr__ = (
        # content: b"startline"
        ("startline", None, None),
        # content: [("name", "value"), ...]
        ("hdr", None, HTTPHeader),
        ("sep", "2s", b"\r\n")
    )

    def _dissect(self, buf):
        # requestline: [method] [uri] [version] eg GET / HTTP/1.1
        # responseline: [version] [status] [reason] eg HTTP/1.1 200 OK
        # logger.debug("Full HTTP: %s", buf)
        # Request/responseline is mendatory to parse header
        if not PROG_STARTLINE.match(buf):
            self.sep = None
            return 0

        try:
            bts_header, bts_body = split_headbody(buf, maxsplit=1)
        # logger.debug("Header: %s\nBody: %s", bts_header, bts_body)
        except ValueError:
            # logger.debug("no startline/header present")
            # deactivate separator
            self.sep = None
            # assume this is part of a bigger (splittet) HTTP-message: no header/only body
            return 0

        try:
            startline, bts_header = split_header(bts_header, maxsplit=1)
        except ValueError:
            # logger.debug("just startline: %r, hdr length=%d" % (bts_header, len(bts_header) + 4))
            # bts_header was something like "HTTP/1.1 123 status" (\r\n\r\n previously removed)
            self.startline = bts_header + b"\r\n"
            # self._init_triggerlist("hdr", b"", lambda _: [])
            return len(bts_header) + 4  # startline + 2 (CR NL) + 0 (header) + 2 (sep: CR NL) + 0 (body)

        self.startline = startline + b"\r\n"
        # bts_header = hdr1\r\nhdr2 -> hdr1\r\nhdr2\r\n
        self._init_triggerlist("hdr", bts_header + b"\r\n", self.__parse_header)

        # logger.debug("startline: %s" % self.startline)
        # logger.debug("hdr: %s" % self.hdr)
        # logger.debug("bts_header: %s" % (bts_header+b"\r\n"))
        # logger.debug("sep: %s" % self.sep)

        # logger.debug(len(startline+b"\r\n") + len(bts_header+b"\r\n\r\n"))
        # logger.debug("lengths head/body: %d %d" % (len(buf), len(bts_body)))
        # logger.debug(buf[:len(buf) - len(bts_body)])
        # HEADER + "\r\n" + BODY -> newline is part of the header
        return len(buf) - len(bts_body)

    @staticmethod
    def __parse_header(buf):
        # logger.debug("parsing header: %s", buf)
        header = []
        lines = split_header(buf)

        for line in lines:
            # logger.debug("checking line: %s", line)
            if len(line) == 0:
                break
            try:
                key, val = split_keyval(line, 1)
                header.append((key, val))
            except:
                # Not a "key: value" line
                logger.warning("Invalid HTTP line: %s", line)
                header.append(line)

        return header



"""Domain Name System."""
import logging

from packetracer import packetracer, triggerlist
from packetracer.packetracer_meta import FIELD_FLAG_AUTOUPDATE
from packetracer.structcbs import unpack_H, unpack_HHHH

logger = logging.getLogger("packetracer")

DNS_Q = 0
DNS_R = 1

# Opcodes
DNS_QUERY = 0
DNS_IQUERY = 1
DNS_STATUS = 2
DNS_NOTIFY = 4
DNS_UPDATE = 5

# Flags
DNS_AN = 0x8000  # this is a response
DNS_CD = 0x0010  # checking disabled
DNS_AD = 0x0020  # authenticated data
DNS_Z = 0x0040  # unused
DNS_RA = 0x0080  # recursion available
DNS_RD = 0x0100  # recursion desired
DNS_TC = 0x0200  # truncated
DNS_AA = 0x0400  # authoritative answer

# Response codes
DNS_RCODE_NOERR = 0
DNS_RCODE_FORMERR = 1
DNS_RCODE_SERVFAIL = 2
DNS_RCODE_NXDOMAIN = 3
DNS_RCODE_NOTIMP = 4
DNS_RCODE_REFUSED = 5
DNS_RCODE_YXDOMAIN = 6
DNS_RCODE_YXRRSET = 7
DNS_RCODE_NXRRSET = 8
DNS_RCODE_NOTAUTH = 9
DNS_RCODE_NOTZONE = 10

# RR types
DNS_A = 1
DNS_NS = 2
DNS_CNAME = 5
DNS_SOA = 6
DNS_WKS = 11
DNS_PTR = 12
DNS_HINFO = 13
DNS_MINFO = 14
DNS_MX = 15
DNS_TXT = 16
DNS_RP = 17
DNS_SIG = 24
DNS_GPOS = 27
DNS_AAAA = 28
DNS_LOC = 29
DNS_SRV = 33
DNS_NAPTR = 35
DNS_KX = 36
DNS_CERT = 37
DNS_DNAME = 39
DNS_DS = 43
DNS_SSHFP = 44
DNS_IPSECKEY = 45
DNS_RRSIG = 46
DNS_NSEC = 47
DNS_DNSKEY = 48
DNS_DHCID = 49
DNS_NSEC3 = 50
DNS_NSEC3PARAM = 51
DNS_TLSA = 52
DNS_SPF = 99
DNS_TKEY = 249
DNS_TSIG = 250
DNS_IXFR = 251
DNS_AXFR = 252
DNS_CAA = 257
DNS_TA = 32768
DNS_DLV = 32769

# RR classes
DNS_IN = 1
DNS_CHAOS = 3
DNS_HESIOD = 4
DNS_ANY = 255


def get_bts_for_msg_compression(tl_packet):
    """return -- header bytes of DNS or b"" """
    # DNS.Triggestlist[sub] -> sub._triggelistpacket_parent == DNS
    if tl_packet._triggelistpacket_parent is not None:
        return tl_packet._triggelistpacket_parent.header_bytes
    return b""


class DNS(packetracer.Packet):
    __hdr__ = (
        ("id", "H", 0x1234),
        ("flags", "H", DNS_AD | DNS_RD),
        ("questions_amount", "H", 0, FIELD_FLAG_AUTOUPDATE),
        ("answers_amount", "H", 0, FIELD_FLAG_AUTOUPDATE),
        ("authrr_amount", "H", 0, FIELD_FLAG_AUTOUPDATE),
        ("addrr_amount", "H", 0, FIELD_FLAG_AUTOUPDATE),
        ("queries", None, triggerlist.TriggerList),
        ("answers", None, triggerlist.TriggerList),
        ("auths", None, triggerlist.TriggerList),
        ("addrecords", None, triggerlist.TriggerList)
    )

    class Query(packetracer.Packet):
        """DNS question."""
        __hdr__ = (
            ("name", None, b"\x03www\x04test\x03com\x00"),
            ("type", "H", DNS_A),
            ("cls", "H", DNS_IN)
        )

        name_s = packetracer.get_property_dnsname("name", cb_mc_bytes=get_bts_for_msg_compression)

        def _dissect(self, buf):
            q_end = DNS.get_dns_length(buf)
            self.name = buf[:q_end]
            # logger.debug("val / format: %s %s" % (self._name, self._name_format))
            return len(buf)  # name (including 0) + type + cls

    class Answer(packetracer.Packet):
        """DNS resource record."""
        __hdr__ = (
            ("name", None, b"\xc0\x0c"),
            ("type", "H", DNS_A),
            ("cls", "H", DNS_IN),
            ("ttl", "I", 180),
            ("dlen", "H", 4),  # length of the next field
            ("address", None, b"1234")  # eg IPv4
        )

        name_s = packetracer.get_property_dnsname("name", cb_mc_bytes=get_bts_for_msg_compression)

        def _get_address_readable(self):
            if self.type == DNS_A:
                return packetracer.ip4_bytes_to_str(self.address)
            elif self.type == DNS_AAAA:
                return packetracer.ip6_bytes_to_str(self.address)
            elif self.type == DNS_CNAME:
                return packetracer.dns_name_decode(self.address)
            else:
                # Unknown type
                return ""

        address_s = property(_get_address_readable)

        def _dissect(self, buf):
            name_end = DNS.get_dns_length(buf)
            self.name = buf[:name_end]
            # needed set format
            start = name_end + 8
            addr_len = unpack_H(buf[start:start + 2])[0]
            self.address = buf[start + 2:start + 2 + addr_len]
            # logger.debug("address: %s" % self.address)
            return start + 2 + addr_len

    class Auth(packetracer.Packet):
        """Auth, generic type."""
        __hdr__ = (
            ("name", "H", 0),
            ("type", "H", 0),
            ("cls", "H", 0),
            ("ttl", "I", 0),
            ("dlen", "H", 0),  # length of the rest of header: server + x, x becmoes body content
            ("server", None, b"\x03www\x04test\x03com\x00")
        )

        server_s = packetracer.get_property_dnsname("server", cb_mc_bytes=get_bts_for_msg_compression)

        def _dissect(self, buf):
            # Needed set format
            # Find server name by 0-termination
            off_end = buf.find(b"\x00", 12)
            if off_end == -1:
                off_end = len(buf)
            else:
                off_end += 1
            self.server = buf[12: off_end]
            # logger.debug("server: %s" % self.server)

            return off_end

    class AuthSOA(packetracer.Packet):
        """
        Auth type SOA.
        Not used atm
        """
        __hdr__ = (
            ("name", "H", 0),
            ("type", "H", 0),
            ("cls", "H", 0),
            ("ttl", "I", 0),
            ("dlen", "H", 0),
            ("name2", None, b"\x03www\x04test\x03com\x00"),
            ("mailbox", None, b"\x03www\x04test\x03com\x00"),
            ("pserver", "H", 0),
            ("mbox", "H", 0),
            ("serial", "H", 0),
            ("refresh", "H", 0),
            ("retry", "H", 0),
            ("expire", "H", 0),
            ("minttl", "H", 0)
        )

        name_s = packetracer.get_property_dnsname("name", get_bts_for_msg_compression)
        mailbox_s = packetracer.get_property_dnsname("mailbox", get_bts_for_msg_compression)

        def _dissect(self, buf):
            # set format
            # find server name by 0-termination
            idx = buf.find(b"\x00", 12)
            # logger.debug(buf[12: idx+1])
            # don't add trailing \0
            self.name = buf[12: idx + 1]
            # logger.debug("name: %s" % buf[idx + 1: -14])
            self.mailbox = buf[idx + 1: -14]
            return len(buf)

    class AddRecord(packetracer.Packet):
        """DNS additional records."""
        __hdr__ = (
            ("name", "H", 0),
            ("type", "H", 0x0001),
            ("clz", "H", 0x0001),
            ("ts", "I", 0),
            ("dlen", "H", 0),
            ("addr", None, b"\x01\x02\x03\x04")
        )

        def _dissect(self, buf):
            # logger.debug(buf[0: idx+1])
            self.addr = buf[12:]
            # logger.debug("addr: %s" % self.addr)
            return len(buf)

    class AddRecordRoot(packetracer.Packet):
        """DNS additional records."""
        __hdr__ = (
            ("name", "B", 0),
            ("type", "H", 0x0001),
            ("udpsize", "H", 0x0001),
            ("rcode", "B", 0),
            ("v", "B", 0),
            ("z", "H", 0),
            ("dlen", "H", 0)
        )

    @staticmethod
    def get_dns_length(bts):
        """
        return -- length of DNS name including terminating 0 (if present)
        """
        off = 0

        while off < len(bts):
            # check for pointer
            if bts[off] == 0xC0:
                # logger.debug("Found pointer at %d", off)
                return off + 2
            # found terminating 0
            elif bts[off] == 0x00:
                # logger.debug("Found 0 byte at %d", off)
                return off + 1
            off += bts[off] + 1

        return 0

    def _dissect(self, buf):
        # unpack basic data to get things done
        quests_amount, ans_amount, authserver_amount, addreq_amount = unpack_HHHH(buf[4:12])
        # Sanity check: assume max amount of 50 addresses
        if quests_amount > 50 or ans_amount > 50 or authserver_amount > 50 or addreq_amount > 50:
            raise Exception("Adress count too high, invalid packet")
        off = 12

        # TODO: use lazy dissect, dns seems to be too shitty for this
        #
        # parse queries
        #
        # logger.debug(">>> parsing questions: %d" % quests_amount)
        while quests_amount > 0 and off < len(buf):
            # find name by 0-termination
            q_end = off + DNS.get_dns_length(buf[off:]) + 4
            # logger.debug("name is: %s" % buf[off: q_end-4])
            # logger.debug("Query is: %s" % buf[off: q_end])
            # logger.debug(len(buf[off: q_end]))
            q = DNS.Query(buf[off: q_end])
            q.dns_bytes = buf
            # logger.debug("query is following..")
            # logger.debug("Query: %s" % q)
            # logger.debug("query name format: %s" % q._name_format)
            self.queries.append(q)
            off = q_end
            quests_amount -= 1

        #
        # parse answers
        #
        # logger.debug(">>> parsing answers: %d" % ans_amount)
        while ans_amount > 0 and off < len(buf):
            # find name by label/0-termination
            # DNS name:x + type:2 + class:2 + ttl:4
            a_end = off + DNS.get_dns_length(buf[off:]) + 2 + 2 + 4
            # logger.debug("name is: %s" % buf[off: a_end-8])
            dlen = unpack_H(buf[a_end: a_end + 2])[0]
            # logger.debug("dlen: %d", dlen)
            # dlen header: 2 + dlen
            a_end += (2 + dlen)
            # logger.debug("Answer is: %r" % buf[off: a_end])
            a = DNS.Answer(buf[off: a_end])
            a.dns_bytes = buf
            # logger.debug("Answer: %s" % a)
            self.answers.append(a)
            off = a_end
            ans_amount -= 1

        #
        # parse authorative servers
        #
        # logger.debug(">>> parsing authorative servers: %d" % authserver_amount)
        while authserver_amount > 0 and off < len(buf):
            dlen = unpack_H(buf[off + 10: off + 12])[0]
            authlen = 12 + dlen
            # logger.debug("Auth: %r" % buf[off: off + authlen])
            a = DNS.Auth(buf[off: off + authlen])
            a.dns_bytes = buf

            # logger.debug("Auth server: %s" % a)
            self.auths.append(a)
            off += authlen
            authserver_amount -= 1

        #
        # parse additional requests
        #
        # logger.debug(">>> parsing additional records: %d" % addreq_amount)
        while addreq_amount > 0 and off < len(buf):
            if buf[off: off + 3] == b"\x00\x00\x29":
                a = DNS.AddRecordRoot(buf[off: off + 11])
                off += 11
            else:
                # logger.debug(buf[idx:])
                # logger.debug(buf[off:])
                # logger.debug("data length via: %r" % buf[idx + 9: idx + 11])
                dlen = unpack_H(buf[off + 10: off + 10 + 2])[0]
                # logger.debug("AddRecord: %s" % buf[off: off + 12 + dlen])
                a = DNS.AddRecord(buf[off: off + 12 + dlen])
                # logger.debug("Additional Record: %s" % a)
                off += 12 + dlen
            self.addrecords.append(a)
            addreq_amount -= 1

        # logger.debug("dns: %s" % self)
        return off

    def _update_fields(self):
        if self._header_changed:
            # logger.debug("updating lenghts")
            # Avoid lazy dissect by checking for [b"bytes", dissect_callback]
            # First assigning to length will trigger _unpack(...)
            if self.questions_amount_au_active and self._queries.__class__ is not list:
                self.questions_amount = len(self.queries)
            if self.answers_amount_au_active and self._answers.__class__ is not list:
                self.answers_amount = len(self.answers)
            if self.authrr_amount_au_active and self._auths.__class__ is not list:
                self.authrr_amount = len(self.auths)
            if self.addrr_amount_au_active and self._addrecords.__class__ is not list:
                self.addrr_amount = len(self.addrecords)

    # logger.debug("finished updating lengths")

    def get_resolved_addresses(self):
        ret = {}

        if self.answers_amount == 0 or self.questions_amount == 0:
            return ret
        """
        if self.questions_amount > 1:
            # Only one question for now
            logger.warning("More than 1 question")
            logger.warning(self.queries)
        """
        question_dns = self.queries[0].name_s

        for answer in self.answers:
            # Assume answer for first query
            if answer.name == b"\xc0\x0c" and answer.type in [DNS_A, DNS_AAAA]:
                ret[answer.address_s] = question_dns[:-1]

        return ret



import logging

logger = logging.getLogger("packetracer")
from_bytes = int.from_bytes


def _get_der_tlv(der_bts):
    off = 0
    tagstart = off
    # logger.debug("tagstart: %d" % tagstart)
    # is_hightag = False
    is_primitive = (der_bts[off] & 0x20) == 0
    # logger.debug("primitive: %r" % is_primitive)

    # high tag number form
    if (der_bts[off] & 0x1F) == 0x1F:
        # logger.debug("got high tag form")
        # is_hightag = True
        off += 1

        while (der_bts[off] & 0x80) != 0:
            off += 1
    off += 1

    lenstart = off
    # logger.debug("lenstart: %d" % lenstart)
    vlen = der_bts[off]
    is_lenshort = (vlen & 0x80) == 0

    if not is_lenshort:
        len_octets = vlen & 0x7F
        lenbts = der_bts[lenstart + 1: lenstart + 1 + len_octets]
        vlen = from_bytes(lenbts, byteorder="big", signed=False)
        # logger.debug("length longform, bytes: %r (%d) = %d" % (lenbts, len_octets, vlen))
        off += len_octets
    off += 1
    valuestart = off
    return lenstart - tagstart, valuestart - lenstart, vlen, is_primitive


def decode_der(der_bts, result=None, extract_cb=None, _firstrun=True, _level=0):
    off = 0
    end = len(der_bts)

    if _firstrun:
        # logger.debug(">>> first run")
        try:
            taglen, lenlen, vlen, prim = _get_der_tlv(der_bts)
        except IndexError:
            result.append(None)
            return
        end = taglen + lenlen + vlen

    while off < end:
        try:
            taglen, lenlen, vlen, prim = _get_der_tlv(der_bts[off:])
        except IndexError:
            result.append(None)
            break

        value = der_bts[off + taglen + lenlen: off + taglen + lenlen + vlen]

        if not prim:
            result_tmp = []
            # logger.debug("sub calling: %d %d %d %d %r" % (off, taglen, lenlen, vlen, _firstrun))
            try:
                decode_der(value, result=result_tmp, extract_cb=extract_cb, _firstrun=False, _level=_level + 1)
            except:
                # None indicates decoding error
                result_tmp = [None]
            value = result_tmp

        result.append([
            der_bts[off: off + taglen],
            der_bts[off + taglen: off + taglen + lenlen],
            value]
        )
        off += (taglen + lenlen + vlen)

    try:
        extract_cb(result)
    except:
        # logger.exception(ex)
        pass



"""
Link Layer Discovery Protocol
IEEE 802.1AB
DCB eXchange protocol
IEEE 802.1Qaz
"""
from packetracer import packetracer, triggerlist
from packetracer.packetracer import (mac_str_to_bytes, mac_bytes_to_str,
                                     ip4_str_to_bytes, ip4_bytes_to_str,
                                     ip6_str_to_bytes, ip6_bytes_to_str)
from packetracer.packetracer_meta import FIELD_FLAG_AUTOUPDATE
from packetracer.structcbs import pack_B, unpack_B, unpack_H, unpack_I

# Mandatory TLV fields length in bytes
# of two field type(7-bit) and length(9-bit)
TLV_HEADER_LEN = 2
TYPE_FIELD_BITS = 7
LENGTH_FIELD_BITS = 9
TYPE_MASK = 0xFE00
LENGTH_MASK = 0x01FF

# Organizationally specific TLV fields length in bytes
# of two field OUI(24-bit) and subtype(8-bit)
ORG_SPEC_TYPE = 127
ORG_SPEC_HEADER_LEN = 4
OUI_MASK = 0xFFFFFF00
# length of TLV Subtype field
SUBTYPE_LEN_BYTE = 1
SUBTYPE_LEN_BITS = 8
SUBTYPE_MASK = 0x000000FF

# Convenient access for value field in Chassis TLV
GET_CHASSIS_TLV_SUBTYPES = {4: mac_bytes_to_str, 5: ip4_bytes_to_str}
SET_CHASSIS_TLV_SUBTYPES = {4: mac_str_to_bytes, 5: ip4_str_to_bytes}
# Convenient access for value field in Port TLV
GET_PORT_TLV_SUBTYPES = {3: mac_bytes_to_str, 4: ip4_bytes_to_str}
SET_PORT_TLV_SUBTYPES = {3: mac_str_to_bytes, 4: ip4_str_to_bytes}
# Convenient access for Management Address
GET_ADDRESS_SUBTYPE = {1: ip4_bytes_to_str, 2: ip6_bytes_to_str, 6: mac_bytes_to_str}
SET_ADDRESS_SUBTYPE = {1: ip4_str_to_bytes, 2: ip6_str_to_bytes, 6: mac_str_to_bytes}


def get_property_tlv_type():
    """Create a get/set-property for type field."""
    return property(
        lambda obj: (obj.type_len & TYPE_MASK) >> LENGTH_FIELD_BITS,
        lambda obj, val: obj.__setattr__("type_len",
                                         (obj.type_len & ~TYPE_MASK) | (val << LENGTH_FIELD_BITS)),
    )


def get_property_tlv_len():
    """Create a get/set-property for length field."""
    return property(
        lambda obj: obj.type_len & LENGTH_MASK,
        lambda obj, val: obj.__setattr__("type_len", (obj.type_len & TYPE_MASK) | val)
    )


def get_property_tlv_oui():
    """Create a get/set-property for OUI field."""
    return property(
        lambda obj: (obj.oui_subtype & OUI_MASK) >> 8,
        lambda obj, val: obj.__setattr__("oui_subtype", (obj.oui_subtype & ~OUI_MASK) | (val << 8))
    )


def get_property_tlv_subtype():
    """Create a get/set-property for subtype field."""
    return property(
        lambda obj: obj.oui_subtype & SUBTYPE_MASK,
        lambda obj, val: obj.__setattr__("oui_subtype",
                                         (obj.oui_subtype & OUI_MASK) | val),
    )


def get_property_to_convert_8_bytes_to_list(var):
    """Create a get/set-property to convert 8 bytes field to list(decimal representation)."""
    return property(
        lambda obj: [unpack_B(x)[0] for x in obj.__getattribute__(var)],
        lambda obj, val: obj.__setattr__(var, [pack_B(x) for x in val]),
    )


def get_property_to_convert_4_bytes_to_list(var):
    """Create a get/set-property to convert 4 bytes field to list(decimal representation)."""
    return property(
        lambda obj: [(obj.__getattribute__(var) >> x) & 0xF for x in reversed(range(0, 32, 4))],
        lambda obj, val: obj.__setattr__(var,
                                         sum([item << bits for item, bits in zip(val, reversed(range(0, 32, 4)))])),
    )


def get_property_to_convert_1_byte_to_list(var):
    """Create a get/set-property to convert 1 byte field to list(bit representation)."""
    return property(
        # lambda obj: [(obj.__getattribute__(var) >> x) & 1 for x in reversed(range(8))],
        lambda obj: [int(bstr) for bstr in bin(obj.__getattribute__(var))[2:]],
        # lambda obj, val: obj.__setattr__(var, int("".join(map(str, val)), 2)),
        lambda obj, val: obj.__setattr__(var, int("".join(["%d" % bint for bint in val]), 2)),
    )


class LLDP(packetracer.Packet):
    __hdr__ = (
        ("tlvlist", None, triggerlist.TriggerList),
    )

    def _dissect(self, buf):
        self._init_triggerlist("tlvlist", buf, self.__parse_tlv)
        return len(buf)

    @staticmethod
    def __parse_tlv(buf):
        """Parse LLDP TLVs and return them as list."""
        _, clz_bts_list = count_and_dissect_tlvs(buf)
        tlvlist = []

        for clz, bts in clz_bts_list:
            tlvlist.append(clz(bts))

        return tlvlist


class LLDPGeneric(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 0, FIELD_FLAG_AUTOUPDATE),
        ("value", None, b""),
    )

    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()

    def _dissect(self, buf):
        self.value = buf[TLV_HEADER_LEN:]
        return len(buf)

    def _update_fields(self):
        if self._changed() and self.type_len_au_active:
            self.tlv_len = len(self) - TLV_HEADER_LEN


class LLDPDUEnd(packetracer.Packet):
    __hdr__ = (
        ("tlv_type", "B", 0),
        ("tlv_len", "B", 0),
    )


class LLDPChassisId(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 512, FIELD_FLAG_AUTOUPDATE),  # type(1)
        ("subtype", "B", 0),
        ("value", None, b""),
    )
    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()

    def __get_value(self):
        if self.subtype not in GET_CHASSIS_TLV_SUBTYPES.keys():
            return self.value.decode()
        return GET_CHASSIS_TLV_SUBTYPES.get(self.subtype)(self.value)

    def __set_value(self, value):
        if self.subtype not in SET_CHASSIS_TLV_SUBTYPES.keys():
            self.value = value
        else:
            self.value = SET_CHASSIS_TLV_SUBTYPES.get(self.subtype)(value)

    value_s = property(__get_value, __set_value)

    def _dissect(self, buf):
        self.value = buf[TLV_HEADER_LEN + SUBTYPE_LEN_BYTE:]
        return len(buf)

    def _update_fields(self):
        if self._changed() and self.type_len_au_active:
            self.tlv_len = len(self) - TLV_HEADER_LEN


class LLDPPortId(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 1024, FIELD_FLAG_AUTOUPDATE),  # type(2)
        ("subtype", "B", 0),
        ("value", None, b""),
    )
    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()

    def __get_value(self):
        if self.subtype not in GET_PORT_TLV_SUBTYPES.keys():
            return self.value.decode()
        return GET_PORT_TLV_SUBTYPES.get(self.subtype)(self.value)

    def __set_value(self, value):
        if self.subtype not in SET_PORT_TLV_SUBTYPES.keys():
            self.value = value
        else:
            self.value = SET_PORT_TLV_SUBTYPES.get(self.subtype)(value)

    value_s = property(__get_value, __set_value)

    def _dissect(self, buf):
        self.value = buf[TLV_HEADER_LEN + SUBTYPE_LEN_BYTE:]
        return len(buf)

    def _update_fields(self):
        if self._changed() and self.type_len_au_active:
            self.tlv_len = len(self) - TLV_HEADER_LEN


class LLDPTTL(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 1538),  # type(3), length(2)
        ("seconds", "H", 0),
    )
    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()


class LLDPPortDescription(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 2048, FIELD_FLAG_AUTOUPDATE),  # type(4)
        ("value", None, b""),
    )

    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()

    def _dissect(self, buf):
        self.value = buf[TLV_HEADER_LEN:]
        return len(buf)

    def _update_fields(self):
        if self._changed() and self.type_len_au_active:
            self.tlv_len = len(self) - TLV_HEADER_LEN


class LLDPSystemName(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 2560, FIELD_FLAG_AUTOUPDATE),  # type(5)
        ("value", None, b""),
    )

    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()

    def _dissect(self, buf):
        self.value = buf[TLV_HEADER_LEN:]
        return len(buf)

    def _update_fields(self):
        if self._changed() and self.type_len_au_active:
            self.tlv_len = len(self) - TLV_HEADER_LEN


class LLDPSystemDescription(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 3072, FIELD_FLAG_AUTOUPDATE),  # type(6)
        ("value", None, b""),
    )

    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()

    def _dissect(self, buf):
        self.value = buf[TLV_HEADER_LEN:]
        return len(buf)

    def _update_fields(self):
        if self._changed() and self.type_len_au_active:
            self.tlv_len = len(self) - TLV_HEADER_LEN


class LLDPSystemCapabilities(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 3588),  # type(7), length(4)
        ("capabilities", "H", 0),
        ("enabled", "H", 0),
    )
    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()


class LLDPManagementAddress(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 4096, FIELD_FLAG_AUTOUPDATE),
        # contains the length of the addrsubtype(1 byte) + addrval(1-31 bytes) fields
        ("addrlen", "B", 2, FIELD_FLAG_AUTOUPDATE),
        ("addrsubtype", "B", 0),
        ("addrval", None, b"\x00"),
        # contains one of subtypes {1: "Unknown", 2: "ifIndex", 3: "System Port Number"}
        ("ifsubtype", "B", 1),
        ("ifnumber", "I", 0),
        ("oidlen", "B", 0, FIELD_FLAG_AUTOUPDATE),
        ("oid", None, b""),
    )

    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()

    def __get_addrval(self):
        if self.addrsubtype not in GET_ADDRESS_SUBTYPE.keys():
            return ""
        return GET_ADDRESS_SUBTYPE.get(self.addrsubtype)(self.addrval)

    def __set_addrval(self, value):
        if self.addrsubtype not in SET_ADDRESS_SUBTYPE.keys():
            self.addrval = value
        else:
            self.addrval = SET_ADDRESS_SUBTYPE.get(self.addrsubtype)(value)

    addrval_s = property(__get_addrval, __set_addrval)

    def _dissect(self, buf):
        addrlen = unpack_B(buf[TLV_HEADER_LEN: TLV_HEADER_LEN + 1])[0]
        addrval_position = TLV_HEADER_LEN + 2
        self.addrval = buf[addrval_position: addrval_position + addrlen - 1]
        oidlen_postion = addrval_position + addrlen + 4
        oidlen = unpack_B(buf[oidlen_postion: oidlen_postion + 1])[0]
        if oidlen:
            self.oid = buf[oidlen_postion + 1: oidlen_postion + 1 + oidlen]
        return len(buf)

    def _update_fields(self):
        if self._changed():
            if self.type_len_au_active:
                self.tlv_len = len(self) - TLV_HEADER_LEN
            if self.addrlen_au_active:
                self.addrlen = len(self.addrval) + SUBTYPE_LEN_BYTE
            if self.oidlen_au_active:
                self.oidlen = len(self.oid)


class LLDPOrgSpecGeneric(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 65024, FIELD_FLAG_AUTOUPDATE),  # type(127)
        ("oui_subtype", "I", 0),
        ("value", None, b""),
    )

    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()
    oui = get_property_tlv_oui()
    subtype = get_property_tlv_subtype()

    def _dissect(self, buf):
        self.value = buf[TLV_HEADER_LEN + ORG_SPEC_HEADER_LEN:]
        return len(buf)

    def _update_fields(self):
        if self._changed() and self.type_len_au_active:
            self.tlv_len = len(self) - TLV_HEADER_LEN


class LLDPDot1PortVlanId(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 65030),  # type(127), length(6)
        ("oui_subtype", "I", 8438273),  # OUI(00-80-C2), subtype(0x01)
        ("vlan", "H", 1),
    )

    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()
    oui = get_property_tlv_oui()
    subtype = get_property_tlv_subtype()


class DCBXCongestionNotification(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 65030),  # type(127), length(6)
        ("oui_subtype", "I", 8438280),  # OUI(00-80-C2), subtype(0x08)
        ("cnpv", "B", 0),
        ("ready", "B", 0),
    )

    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()
    oui = get_property_tlv_oui()
    subtype = get_property_tlv_subtype()
    cnpv_list = get_property_to_convert_1_byte_to_list("cnpv")
    ready_list = get_property_to_convert_1_byte_to_list("ready")


class DCBXConfiguration(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 65049),  # type(127), length(25)
        ("oui_subtype", "I", 8438281),  # OUI(00-80-C2), subtype(0x09)
        ("w_cbs_maxtc", "B", 0),  # Field contains Willing-1bit, CBS-1bit, Reserved-3bit,  Max TCs-3bit
        ("priority", "I", 0),  # Field represents list of 8 items where one item = 4 bits
        ("tcbandwith", None, triggerlist.TriggerList),
        ("tsaassigment", None, triggerlist.TriggerList),
    )

    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()
    oui = get_property_tlv_oui()
    subtype = get_property_tlv_subtype()
    tcbandwith_list = get_property_to_convert_8_bytes_to_list("tcbandwith")
    tsaassigment_list = get_property_to_convert_8_bytes_to_list("tsaassigment")
    priority_list = get_property_to_convert_4_bytes_to_list("priority")

    def __get_w(self):
        return (self.w_cbs_maxtc & 0x80) >> 7

    def __set_w(self, value):
        self.w_cbs_maxtc = (self.w_cbs_maxtc & ~0x80) | (value << 7)

    willing = property(__get_w, __set_w)

    def __get_cbs(self):
        return (self.w_cbs_maxtc & 0x40) >> 6

    def __set_cbs(self, value):
        self.w_cbs_maxtc = (self.w_cbs_maxtc & ~0x40) | (value << 6)

    cbs = property(__get_cbs, __set_cbs)

    def __get_maxtcs(self):
        return self.w_cbs_maxtc & 0x07

    def __set_maxtcs(self, value):
        self.w_cbs_maxtc = self.w_cbs_maxtc & 0xF8 | value

    maxtcs = property(__get_maxtcs, __set_maxtcs)

    def _dissect(self, buf):
        for i in range(11, 19):
            self.tcbandwith.append(buf[i:i + 1])
        for i in range(19, 27):
            self.tsaassigment.append(buf[i:i + 1])
        return len(self)


class DCBXRecommendation(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 65049),  # type(127), length(25)
        ("oui_subtype", "I", 8438282),  # OUI(00-80-C2), subtype(0x10)
        ("reserved", "B", 0),
        ("priority", "I", 0),  # Field represents list of 8 items where one item = 4 bits
        ("tcbandwith", None, triggerlist.TriggerList),
        ("tsaassigment", None, triggerlist.TriggerList),
    )

    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()
    oui = get_property_tlv_oui()
    subtype = get_property_tlv_subtype()
    tcbandwith_list = get_property_to_convert_8_bytes_to_list("tcbandwith")
    tsaassigment_list = get_property_to_convert_8_bytes_to_list("tsaassigment")
    priority_list = get_property_to_convert_4_bytes_to_list("priority")

    def _dissect(self, buf):
        # start from TLV_HEADER_LEN + ORG_SPEC_HEADER_LEN +
        # 1 byte(reserved) + 4 bytes(priority)
        for i in range(11, 19):
            self.tcbandwith.append(buf[i:i + 1])
        for i in range(19, 27):
            self.tsaassigment.append(buf[i:i + 1])
        return len(self)


class DCBXPriorityBasedFlowControlConfiguration(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 65030),  # type(127), length(6)
        ("oui_subtype", "I", 8438283),  # OUI(00-80-C2), subtype(0x11)
        ("w_mbc_pfc", "B", 0),  # Field contains Willing-1bit, MBC-1bit, Reserved-2bit,  PFC cap TCs-4bit
        ("pfcenable", "B", 0),
    )

    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()
    oui = get_property_tlv_oui()
    subtype = get_property_tlv_subtype()
    pfcenable_list = get_property_to_convert_1_byte_to_list("pfcenable")

    def __get_w(self):
        return (self.w_mbc_pfc & 0x80) >> 7

    def __set_w(self, value):
        self.w_mbc_pfc = (self.w_mbc_pfc & ~0x80) | (value << 7)

    willing = property(__get_w, __set_w)

    def __get_mbc(self):
        return (self.w_mbc_pfc & 0x40) >> 6

    def __set_mbc(self, value):
        self.w_mbc_pfc = (self.w_mbc_pfc & ~0x40) | (value << 6)

    mbc = property(__get_mbc, __set_mbc)

    def __get_pfccap(self):
        return self.w_mbc_pfc & 0x0F

    def __set_pfccap(self, value):
        self.w_mbc_pfc = self.w_mbc_pfc & 0xF0 | value

    pfccap = property(__get_pfccap, __set_pfccap)


class DCBXApplicationPriority(packetracer.Packet):
    __hdr__ = (
        ("type_len", "H", 65024, FIELD_FLAG_AUTOUPDATE),  # type(127)
        ("oui_subtype", "I", 8438284),  # OUI(00-80-C2), subtype(0x12)
        ("reserved", "B", 0),
        ("apppriotable", None, triggerlist.TriggerList),
    )

    tlv_type = get_property_tlv_type()
    tlv_len = get_property_tlv_len()
    oui = get_property_tlv_oui()
    subtype = get_property_tlv_subtype()

    def _dissect(self, buf):
        # start from TLV_HEADER_LEN + ORG_SPEC_HEADER_LEN + 1 byte(reserved)
        for i in range(7, len(buf), 3):
            self.apppriotable.append(DCBXApplicationPriorityTable(buf[i:i + 3]))
        return len(self)

    def _update_fields(self):
        if self._changed() and self.type_len_au_active:
            self.tlv_len = len(self) - TLV_HEADER_LEN


class DCBXApplicationPriorityTable(packetracer.Packet):
    __hdr__ = (
        ("priority_sel", "B", 0),  # Field contains Priority-3bits, Reserved-2bits,  Sel-3bits)
        ("protocolid", "H", 0),
    )

    def __get_prio(self):
        return (self.priority_sel & 0xE0) >> 5

    def __set_prio(self, value):
        self.priority_sel = (self.priority_sel & ~0xE0) | (value << 5)

    priority = property(__get_prio, __set_prio)

    def __get_sel(self):
        return self.priority_sel & 0x07

    def __set_sel(self, value):
        self.priority_sel = (self.priority_sel & 0xF8) | value

    sel = property(__get_sel, __set_sel)


LLDP_TLV_CLS = {
    0: LLDPDUEnd,
    1: LLDPChassisId,
    2: LLDPPortId,
    3: LLDPTTL,
    4: LLDPPortDescription,
    5: LLDPSystemName,
    6: LLDPSystemDescription,
    7: LLDPSystemCapabilities,
    8: LLDPManagementAddress
}

LLDP_ORG_SPEC_TLV_CLS = {
    (0x0080C2, 0x01): LLDPDot1PortVlanId,
    (0x0080C2, 0x08): DCBXCongestionNotification,
    (0x0080C2, 0x09): DCBXConfiguration,
    (0x0080C2, 0x0A): DCBXRecommendation,
    (0x0080C2, 0x0B): DCBXPriorityBasedFlowControlConfiguration,
    (0x0080C2, 0x0C): DCBXApplicationPriority,
}


def count_and_dissect_tlvs(buf):
    """
    Count and dissect TLVs. Return length of LLDP layer

    buf -- buffer to dissect
    return -- parsed_bytes_total, [(clz, bts), ...]
    """
    shift = 0
    tlv_type, tlv_len = 1, 1
    clz_bts_list = []

    while (tlv_type | tlv_len) != 0:
        type_and_len = unpack_H(buf[shift:shift + TLV_HEADER_LEN])[0]
        # get tlv length and type
        tlv_type = (type_and_len & TYPE_MASK) >> LENGTH_FIELD_BITS
        tlv_len = type_and_len & LENGTH_MASK

        if tlv_type != ORG_SPEC_TYPE:
            clz = LLDP_TLV_CLS.get(tlv_type, LLDPGeneric)
        else:
            oui_subtype = unpack_I(buf[shift + TLV_HEADER_LEN:shift + ORG_SPEC_HEADER_LEN + TLV_HEADER_LEN])[0]
            oui = (oui_subtype & OUI_MASK) >> SUBTYPE_LEN_BITS
            subtype = oui_subtype & SUBTYPE_MASK
            clz = LLDP_ORG_SPEC_TLV_CLS.get((oui, subtype), LLDPOrgSpecGeneric)
        # get body bytes
        tlv_body = buf[shift: tlv_len + shift + TLV_HEADER_LEN]
        # update shift to begin of next TLV (TLV_HEADER_LEN:2 + content:x)
        shift += TLV_HEADER_LEN + tlv_len
        clz_bts_list.append((clz, tlv_body))

    return shift, clz_bts_list



"""Point-to-Point Protocol."""
import logging

from packetracer import packetracer, triggerlist
# handler
from packetracer.layer3 import ip, ip6
from packetracer.structcbs import unpack_H

logger = logging.getLogger("packetracer")

# http://www.iana.org/assignments/ppp-numbers
PPP_IP = 0x21  # Internet Protocol
PPP_IP6 = 0x57  # Internet Protocol v6

# Protocol field compression
PFC_BIT = 0x01


class PPP(packetracer.Packet):
    __hdr__ = (
        ("p", None, triggerlist.TriggerList),
    )

    __handler__ = {
        PPP_IP: ip.IP,
        PPP_IP6: ip6.IP6
    }

    def _dissect(self, buf):
        offset = 1
        ppp_type = buf[0]

        if buf[0] & PFC_BIT == 0:
            ppp_type = unpack_H(buf[:2])[0]
            offset = 2
            self.p.append(buf[0:2])
        else:
            self.p.append(buf[0:1])
        self._init_handler(ppp_type, buf[offset:])
        return offset



"""
Remote Framebuffer Protocol
# http://www.realvnc.com/docs/rfbproto.pdf
# RFP uses dynamic ports 5900+, we won't auto-decode this!
"""
from packetracer import packetracer

# Client to Server Messages
CLIENT_SET_PIXEL_FORMAT = 0
CLIENT_SET_ENCODINGS = 2
CLIENT_FRAMEBUFFER_UPDATE_REQUEST = 3
CLIENT_KEY_EVENT = 4
CLIENT_POINTER_EVENT = 5
CLIENT_CUT_TEXT = 6

# Server to Client Messages
SERVER_FRAMEBUFFER_UPDATE = 0
SERVER_SET_COLOUR_MAP_ENTRIES = 1
SERVER_BELL = 2
SERVER_CUT_TEXT = 3


class RFB(packetracer.Packet):
    __hdr__ = (
        ("type", "B", 0),
    )


class SetPixelFormat(packetracer.Packet):
    __hdr__ = (
        ("pad", "3s", b"A" * 3),
        ("pixel_fmt", "16s", b"A" * 16)
    )


class SetEncodings(packetracer.Packet):
    __hdr__ = (
        ("pad", "1s", b"A"),
        ("num_encodings", "H", 0)
    )


class FramebufferUpdateRequest(packetracer.Packet):
    __hdr__ = (
        ("incremental", "B", 0),
        ("x_position", "H", 0),
        ("y_position", "H", 0),
        ("width", "H", 0),
        ("height", "H", 0)
    )


class KeyEvent(packetracer.Packet):
    __hdr__ = (
        ("down_flag", "B", 0),
        ("pad", "2s", b"A" * 2),
        ("key", "I", 0)
    )


class PointerEvent(packetracer.Packet):
    __hdr__ = (
        ("button_mask", "B", 0),
        ("x_position", "H", 0),
        ("y_position", "H", 0)
    )


class FramebufferUpdate(packetracer.Packet):
    __hdr__ = (
        ("pad", "1s", b"A" * 1),
        ("num_rects", "H", 0)
    )


class SetColourMapEntries(packetracer.Packet):
    __hdr__ = (
        ("pad", "1s", b"A" * 1),
        ("first_colour", "H", 0),
        ("num_colours", "H", 0)
    )


class CutText(packetracer.Packet):
    __hdr__ = (
        ("pad", "3s", b"A" * 3),
        ("length", "I", 0)
    )



"""
Prism header.
This packet type exists just for convenience. Radiotap should be prefered over prism
because of its superior flexibility. Only use this if there is no support for Radiotap
eg for some Broadcom-Chipsets (stop buying crap man).
"""
import logging

from packetracer import packetracer, triggerlist
# handler
from packetracer.layer12 import ieee80211

logger = logging.getLogger("packetracer")

PRISM_TYPE_80211 = 0
PRISM_DID_RSSI = 0x41400000


class Did(packetracer.Packet):
    __hdr__ = (
        ("id", "I", 0),
        ("status", "H", 0),
        ("len", "H", 0),
        ("value", "I", 0),
    )


class Prism(packetracer.Packet):
    __hdr__ = (
        ("code", "I", 0),
        ("len", "I", 144),
        ("dev", "16s", b"\x00" * 16),
        ("dids", None, triggerlist.TriggerList),
    )

    __handler__ = {
        PRISM_TYPE_80211: ieee80211.IEEE80211
    }

    def _dissect(self, buf):
        off = 24
        # assume 10 DIDs, 24 + 10*12 = 144 bytes prism header
        end = off + 10 * 12

        dids = []

        while off < end:
            did = Did(buf[off:off + 12])
            dids.append(did)
            off += 12

        self.dids.extend(dids)
        self._init_handler(PRISM_TYPE_80211, buf)
        return off



"""
Primaly refer:
	http://www.winpcap.org/ntar/draft/PCAP-DumpFileFormat.html

TODO:
	* Writer class implementation.
	* Options getter/setter implementation.
	* Support nanosecond.
		Investigate the implementation to support multi interface.

Limitation:
	Because the generally considered to only use wireshark,	support header
	referenced to http://wiki.wireshark.org/Development/PcapNg

	Mostly the following limitations
		* Only a single section
		Support:
			+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
			| SHB v1.0  |         Data          |
			+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
		Not support:
			|<-   1st Section   ->|<-   2nd Section   ->| ... |
			+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
			| SHB v1.0  |  Data   | SHB v1.1  |  Data   | ... |
			+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	Wireshark wiki written as "Capture file will have the following
	pcap-ng blocks: SHB, IDB, IDB, IDB, EPB, EPB, ..., ISB, ISB, ISB.".
		* SHB(Section Header Block)
		* IDB(Interface Description Block)
		* EPB(Enhanced Packet Block) <-- this is DUMP PACKET
		* ISB(Interface Statistics Block)
	Thus, created by assuming the following figure of block format:
			+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
			| SHB | IDB | IDB | EPB | EPB | ... | EPB | ISB | ISB |
			+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	( 'A`)< Only file reading...

TODO: generic interface for different reader/writer
"""
import struct

from packetracer import packetracer

# avoid references
unpack = struct.unpack

PCAPNG_IDB = 0x00000001  # Interface Description Block
# PCAPNG_PB  = 0x00000002		# (obsolated) Packet Block
PCAPNG_SPB = 0x00000003  # Simple Packet Block
# PCAPNG_NRB = 0x00000004		# Name Resolution Block
PCAPNG_ISB = 0x00000005  # Interface Statistics Block
PCAPNG_EPB = 0x00000006  # Enhanced Packet Block
PCAPNG_SHB = 0x0A0D0D0A  # Section Header Block

PCAPNG_VERSION_MAJOR = 1
PCAPNG_VERSION_MINOR = 0

BE_MAGIC = 0x1A2B3C4D
LE_MAGIC = 0x4D3C2B1A

OPT_ENDOFOPT = 0
OPT_COMMENT = 1
OPT_IDB_IF_NAME = 2
OPT_IDB_IF_DESC = 3
OPT_IDB_IF_V4ADDR = 4
OPT_IDB_IF_V6ADDR = 5
OPT_IDB_IF_MAC = 6
OPT_IDB_IF_EUI = 7
OPT_IDB_IF_SPEED = 8
OPT_IDB_IF_TSRESOL = 9
OPT_IDB_IF_TZONE = 10
OPT_IDB_IF_FILTER = 11
OPT_IDB_IF_OS = 12
OPT_IDB_IF_FCSLEN = 13
OPT_IDB_IF_TSOFFSET = 14
OPT_ISB_STARTTIME = 2
OPT_ISB_ENDTIME = 3
OPT_ISB_IFRECV = 4
OPT_ISB_IFDROP = 5
OPT_ISB_FILTERACCEPT = 6
OPT_ISB_OSDROP = 7
OPT_ISB_USRDELIV = 8
OPT_SHB_HARDWARE = 2
OPT_SHB_OS = 3
OPT_SHB_USERAPPL = 4

IDB_OPTIONS = {
    1: "opt_comment",
    2: "if_name",
    3: "if_description",
    4: "if_IPv4addr",
    5: "if_IPv6addr",
    6: "if_MACaddr",
    7: "if_EUIaddr",
    8: "if_speed",
    9: "if_tsresol",
    10: "if_tzone",
    11: "if_filter",
    12: "if_os",
    13: "if_fcslen",
    14: "if_tsoffset",
}

ISB_OPTIONS = {
    1: "opt_comment",
    2: "isb_starttime",
    3: "isb_endtime",
    4: "isb_ifrecv",
    5: "isb_ifdrop",
    6: "isb_filteraccept",
    7: "isb_osdrop",
    8: "isb_usrdeliv",
}

SHB_OPTIONS = {
    1: "opt_comment",
    2: "shb_hardware",
    3: "shb_os",
    4: "shb_userappl",
}


def _32bit_alignment(offset, length):
    return (offset + length + 3) & 0xFFFC


class OPT(packetracer.Packet):
    """General option format"""
    __hdr__ = (
        ("code", "H", 0),
        ("length", "H", 0),
    )


class SHB(packetracer.Packet):
    """Section Header Block (mandatory)"""
    __hdr__ = (
        ("type", "I", PCAPNG_SHB),
        ("block_length", "I", 0),
        ("magic", "I", BE_MAGIC),
        ("v_major", "H", PCAPNG_VERSION_MAJOR),
        ("v_minor", "H", PCAPNG_VERSION_MINOR),
        ("section_length", "Q", 0),
    )

    class OPT(OPT):
        pass


class SHBLe(SHB):
    __byte_order__ = "<"

    class OPT(SHB.OPT):
        __byte_order__ = "<"


class IDB(packetracer.Packet):
    """Interface Description Block (mandatory)"""
    __hdr__ = (
        ("type", "I", PCAPNG_IDB),
        ("block_length", "I", 0),
        ("linktype", "H", 0),
        ("reserved", "H", 0),
        ("snaplen", "I", 0),
    )

    class OPT(OPT):
        pass


class IDBLe(IDB):
    __byte_order__ = "<"

    class OPT(IDB.OPT):
        __byte_order__ = "<"


class EPB(packetracer.Packet):
    """Enhanced Packet Block (optional)"""
    __hdr__ = (
        ("type", "I", PCAPNG_EPB),
        ("block_length", "I", 32),
        ("interface_id", "I", 0),
        ("ts_high", "I", 0),
        ("ts_low", "I", 0),
        ("cap_len", "I", 0),
        ("len", "I", 0),
    )

    class OPT(OPT):
        pass


class EPBLe(EPB):
    __byte_order__ = "<"

    class OPT(EPB.OPT):
        __byte_order__ = "<"


class SPB(packetracer.Packet):
    """Simple Packet Block (optional)"""
    __hdr__ = (
        ("type", "I", PCAPNG_SPB),
        ("block_length", "I", 16),
        ("len", "I", 0),
    )


class SPBLe(SPB):
    __byte_order__ = "<"


class ISB(packetracer.Packet):
    """Interface Statistics Block (optional)"""
    __hdr__ = (
        ("type", "I", PCAPNG_ISB),
        ("block_length", "I", 0),
        ("interface_id", "I", 1),
        ("ts_high", "I", 0),
        ("ts_low", "I", 0),
    )

    class OPT(OPT):
        pass


class ISBLe(ISB):
    __byte_order__ = "<"

    class OPT(ISB.OPT):
        __byte_order__ = "<"


# TODO: Writer
class Writer(object):
    def __init__(self):
        pass


class Reader(object):
    def __init__(self,
                 fileobj=None,
                 filename=None):

        self.idbs = []
        self.isbs = []
        self.__block_order__ = ""
        self._IDB = IDB
        self._EPB = EPB
        self._ISB = ISB
        self._SHB = SHB

        # handle source modes
        if fileobj is not None:
            self.__fh = fileobj
        elif filename is not None:
            self.__fh = open(filename, "rb")
        else:
            raise Exception("No fileobject and no filename given..nothing to read!!!")

        # How to parse:
        #
        # 	|--> Parse1       |--> iter               | Parse2 <--|
        # 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # 	| SHB | IDB | IDB | EPB | EPB | ... | EPB | ISB | ISB |
        # 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #
        # 	1. Parse from head and stop at the EPB pointer.
        # 	2. Parse from tail and stop at the EPB pointer
        # 		or not supported header.

        # Parse1
        while 1:
            buf = self.__fh.read(8)
            block_type, block_length = unpack(self.__block_order__ + "2I", buf)
            if block_type == PCAPNG_SHB:
                buf = buf + self.__fh.read(block_length - len(buf))
                self.shb = self._SHB(buf)
                # Endian is decided magic in SHB.
                if self.shb.magic == LE_MAGIC:
                    self.__to_le()
                    self.shb = self._SHB(buf)
                self.shb.opts = self.__unpack_opt(buf, self._SHB)

            elif block_type == PCAPNG_IDB:
                buf = buf + self.__fh.read(block_length - len(buf))
                _idb = self._IDB(buf)
                _idb.opts = self.__unpack_opt(buf, self._IDB)
                self.idbs.append(_idb)

            elif block_type == PCAPNG_EPB:
                self.__iter_pos = self.__fh.tell() - 8
                self.__next__ = self._next_bytes_conversion
                # TODO: Support nanosecond
                self.__resolution_factor = 1000000.0
                break

            else:
                break

        # Parse2
        # 1. Read Block Total Length from tail.
        # 2. Seek reverse the Block Total Length.
        # 3. Same Parse1.
        #
        # 0                   1                   2                   3
        # 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ <---2
        # |                          Block Type                           | 3
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ |
        # |                      Block Total Length                       | V
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # /                          Block Body                           /
        # /          /* variable length, aligned to 32 bits */            /
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |                      Block Total Length                       |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ <---1
        tail_offset = 0
        while 1:
            self.__fh.seek(-1 * (4 + tail_offset), 2)
            block_length = unpack(self.__block_order__ + "I", self.__fh.read(4))[0]
            self.__fh.seek(-1 * block_length, 1)
            buf = self.__fh.read(8)
            block_type, block_length = unpack(self.__block_order__ + "2I", buf)
            tail_offset += block_length

            if block_type == PCAPNG_ISB:
                buf = buf + self.__fh.read(block_length - len(buf))
                _isb = self._ISB(buf)
                _isb.opts = self.__unpack_opt(buf, self._ISB)
                self.isbs.append(_isb)

            else:
                break

    def __to_le(self):
        self.__block_order__ = "<"
        self._IDB = IDBLe
        self._EPB = EPBLe
        self._ISB = ISBLe
        self._SHB = SHBLe

    def __unpack_opt(self, buf, block):
        offset = block._hdr_fmt.size
        opts = []
        while 1:
            opt_hdr = buf[offset:offset + OPT._hdr_fmt.size]
            if not opt_hdr:
                break
            _, length = unpack(self.__block_order__ + "2H", opt_hdr)
            opt = block.OPT(buf[offset:offset + OPT._hdr_fmt.size + length])
            if opt.code == OPT_ENDOFOPT:
                break
            opts.append(opt)
            offset = _32bit_alignment(offset + OPT._hdr_fmt.size, length)
        return opts

    def _next_bytes_conversion(self):
        """
        Standard __next__ implementation. Needs to be a sepearte method to be called by producer.

        return -- (timestamp_microseconds, Enhanced_Packet_Block) for pcap-reader.
            Access DUMP DATA: Enhanced_Packet_Block.data
        """
        buf = self.__fh.read(8)
        if not buf:
            raise StopIteration

        block_type, block_length = unpack(self.__block_order__ + "2I", buf)
        if not block_type == PCAPNG_EPB:
            raise StopIteration

        buf = buf + self.__fh.read(block_length - len(buf))
        _epb = self._EPB(buf)
        _epb.opts = self.__unpack_opt(buf, self._EPB)

        return (((_epb.ts_high << 32) + _epb.ts_low) / self.__resolution_factor, _epb)

    def __iter__(self):
        """
        return -- (timestamp, Enhanced Packet Block) for pcap-reader depending on configuration.
        """
        self.__fh.seek(self.__iter_pos)
        while True:
            try:
                yield self.__next__()
            except StopIteration:
                break



"""Precision Time Protocol v2, IEEE 1588-2008"""
import logging

from packetracer import packetracer
from packetracer.structcbs import pack_Q, unpack_Q

logger = logging.getLogger("packetracer")

PTPv2_TYPE_SYNC = 0x0  # Flags: PTP_TWO_STEP = True
PTPv2_TYPE_DELAY_REQ = 0x1
# IEEE802.1AS is the Audio-Visual Bridging (AVB) profile of the IEEE1588 Precision Time Protocol
PTPv2_TYPE_PATH_DELAY_REQ = 0x2
PTPv2_TYPE_PATH_DELAY_RESP = 0x3

PTPv2_TYPE_FOLLOW_UP = 0x8
PTPv2_TYPE_DELAY_RESP = 0x9
PTPv2_TYPE_PATH_DELAY_RESP_FOLLOWUP = 0xA
PTPv2_TYPE_ANNOUNCE = 0xB  # Flags: PTP_TIMESCALE = True
PTPv2_TYPE_SIGNALLING = 0xC
PTPv2_TYPE_MGMT = 0xD

CTRL_TYPE_SYNC = 0x0
CTRL_TYPE_DELAY_REQ = 0x1
CTRL_TYPE_FOLLOW_UP = 0x2
CTRL_TYPE_DELAY_RESP = 0x3
CTRL_TYPE_MGMT = 0x4
CTRL_TYPE_OTHER = 0x5

TYPES_TS_ACTIVATE = {PTPv2_TYPE_SYNC, PTPv2_TYPE_FOLLOW_UP, PTPv2_TYPE_ANNOUNCE,
                     PTPv2_TYPE_PATH_DELAY_RESP}
TYPES_REQ_PORT_ACTIVATE = {PTPv2_TYPE_PATH_DELAY_RESP, PTPv2_TYPE_PATH_DELAY_RESP_FOLLOWUP}


class PTPv2(packetracer.Packet):
    __hdr__ = (
        ("transport_id", "B", PTPv2_TYPE_DELAY_REQ),
        ("version", "B", 2),
        ("msglen", "H", 0),
        ("subdomains", "B", 0),
        ("rsv", "B", 0),
        ("flags", "H", 0),
        ("corr", "Q", 0),
        ("rsv2", "I", 0),
        ("clockid", "Q", 0),
        ("srcport", "H", 0),
        ("seqid", "H", 0),
        ("ctrl", "B", 0),
        ("log", "B", 0),
        ("tssec_bts", "6s", None),
        ("tsnano", "I", None),
        ("reqclockid", "Q", None),
        ("reqportid", "H", None)
    )

    def __get_transport(self):
        return (self.transport_id & 0xF0) >> 4

    def __set_transport(self, value):
        self.transport_id = ((value & 0xF) << 4) | (self.transport_id & 0xF)

    transport = property(__get_transport, __set_transport)

    def __get_id(self):
        return self.transport_id & 0xF

    def __set_id(self, value):
        self.transport_id = (self.transport_id & 0xF0) | (value & 0xF)

    id = property(__get_id, __set_id)

    def __get_tssec(self):
        return unpack_Q(b"\x00\x00" + self.tssec_bts)[0]

    def __set_tssec(self, value):
        self.tssec_bts = pack_Q(value)[2:]  # (8-2) bytes

    # allows setting tssec as integer in contrast to tssec_bts
    tssec = property(__get_tssec, __set_tssec)

    class Announce(packetracer.Packet):
        __hdr__ = (
            ("utcoff", "H", 0),
            ("prio1", "H", 0),
            ("clockclass", "B", 0),
            ("clockaccuracy", "B", 0),
            ("clockvariance", "H", 0),
            ("prio2", "B", 0),
            ("clockid", "Q", 0),
            ("stepsremoved", "H", 0),
            ("timesource", "B", 0)
        )

    __handler__ = {
        PTPv2_TYPE_ANNOUNCE: Announce
    }

    def _dissect(self, buf):
        header_len = 34
        ptpv2_type = buf[0] & 0xF

        if ptpv2_type in TYPES_TS_ACTIVATE:
            # logger.debug("activating ts fields")
            self.tssec_bts = b"\x00" * 6
            self.tsnano = 0
            header_len += 10

        if ptpv2_type in TYPES_REQ_PORT_ACTIVATE:
            # logger.debug("activating req fields")
            self.reqclockid = 0
            self.reqportid = 0
            header_len += 10
        elif ptpv2_type == PTPv2_TYPE_ANNOUNCE:
            # logger.debug("got announce")
            self._init_handler(ptpv2_type, buf[header_len:])

        if len(buf) < header_len:
            logger.warning("not enough bytes for header: %d < %d", len(buf), header_len)

        return header_len



from packetracer import packetracer
# handler
from packetracer.layer12 import arp
from packetracer.layer3 import ip, ip6
from packetracer.structcbs import unpack_H

LLC_TYPE_IP = 0x0800  # IPv4 protocol
LLC_TYPE_ARP = 0x0806  # address resolution protocol
LLC_TYPE_IP6 = 0x86DD  # IPv6 protocol


class LLC(packetracer.Packet):
    __hdr__ = (
        ("dsap", "B", 0),
        ("ssap", "B", 0),
        ("ctrl", "B", 0),
        ("snap", "5s", b"\x00" * 5),
    )

    __handler__ = {
        LLC_TYPE_IP: ip.IP,
        LLC_TYPE_ARP: arp.ARP,
        LLC_TYPE_IP6: ip6.IP6
    }

    def _dissect(self, buf):
        if buf[0] == 170:  # = 0xAA
            # SNAP is following ctrl
            htype = unpack_H(buf[5:7])[0]
            self._init_handler(htype, buf[8:])
        else:
            # deactivate SNAP
            self.snap = None
        return 8



# Copyright 2012 Google Inc. All rights reserved.

"""
Nicely formatted cipher suite definitions for TLS

A list of cipher suites in the form of CipherSuite objects.
These are supposed to be immutable; don't mess with them.
"""


class CipherSuite(object):
    """
    Encapsulates a cipher suite.

    Members/args:
    * code: two-byte ID code, as int
    * name: as in 'TLS_RSA_WITH_RC4_40_MD5'
    * kx: key exchange algorithm, string
    * auth: authentication algorithm, string
    * encoding: encoding algorithm
    * mac: message authentication code algorithm
    """

    def __init__(self, code, name, kx, auth, encoding, mac):
        self.code = code
        self.name = name
        self.kx = kx
        self.auth = auth
        self.encoding = encoding
        self.mac = mac

    def __repr__(self):
        return 'CipherSuite(%s)' % self.name

    MAC_SIZES = {
        'MD5': 16,
        'SHA': 20,
        'SHA256': 32,  # I guess
    }

    BLOCK_SIZES = {
        'AES_256_CBC': 16,
    }

    @property
    def mac_size(self):
        """In bytes. Default to 0."""
        return self.MAC_SIZES.get(self.mac, 0)

    @property
    def block_size(self):
        """In bytes. Default to 1."""
        return self.BLOCK_SIZES.get(self.encoding, 1)


# master list of CipherSuite Objects
CIPHERSUITES = [
    # not a real cipher suite, can be ignored, see RFC5746
    CipherSuite(0xFF, 'TLS_EMPTY_RENEGOTIATION_INFO', 'NULL', 'NULL', 'NULL', 'NULL'),
    CipherSuite(0x00, 'TLS_NULL_WITH_NULL_NULL', 'NULL', 'NULL', 'NULL', 'NULL'),
    CipherSuite(0x01, 'TLS_RSA_WITH_NULL_MD5', 'RSA', 'RSA', 'NULL', 'MD5'),
    CipherSuite(0x02, 'TLS_RSA_WITH_NULL_SHA', 'RSA', 'RSA', 'NULL', 'SHA'),
    # not sure I got the kx/auth thing right.
    CipherSuite(0x0039, 'TLS_DHE_RSA_WITH_AES_256_CBC_SHA', 'DHE', 'RSA', 'AES_256_CBC', 'SHA'),
    CipherSuite(0xFFFF, 'UNKNOWN_CIPHER', '', '', '', '')
]

# BY_CODE = decoder_dict((cipher.code, cipher) for cipher in CIPHERSUITES)
# BY_NAME = decoder_dict((suite.name, suite) for suite in CIPHERSUITES)
# NULL_SUITE = BY_CODE[0x00]



"""
Logic to build state machines. Borrowed from Scapy's Automata concept.
"""
import collections
import logging
import threading

logger = logging.getLogger("packetracer")

STATE_TYPE_BEGIN = 0
STATE_TYPE_INTERM = 1  # default
STATE_TYPE_END = 2


class TimedCallback(threading.Thread):
    def __init__(self):
        self._obj = None
        self._is_running = True
        self._cb = None
        # assume this will never trigger
        self._timeout = 9999999
        self._event = threading.Event()
        super().__init__()
        self.start()

    def run(self):
        # logger.debug("starting cb iterator")

        while self._is_running:
            # logger.debug("cb: next round")
            self._event.clear()
            self._event.wait(timeout=self._timeout)

            # wait was interrupted
            if self._event.is_set():
                continue

            if self._cb is not None:
                # logger.debug("executing timeout cb")
                self._cb(self._obj)

    # logger.debug("cb iterator finished")

    def retrigger(self, obj, timeout, cb):
        self._obj = obj
        self._timeout = timeout
        self._cb = cb
        self._event.set()

    def set_inactive(self):
        self._cb = None
        self._timeout = 9999999
        self._event.set()

    def stop(self):
        self._is_running = False
        self._event.set()


_cb_threads = collections.defaultdict(TimedCallback)


def sm_state(state_type=STATE_TYPE_INTERM, timeout=None, timeout_cb=None):
    def gen(old_f):
        if timeout is not None and timeout_cb is None:
            logger.warning(
                "timeout set to %d but no timeout action for %r",
                timeout,
                old_f.__name__)

        # replace with new method to store state infos
        def new_f(self, *args, **kwds):
            # end of function (state) -> clear old one
            # logger.debug("getting cb class via %r", self.__class__)
            cb_thread = _cb_threads[self.__class__]
            cb_thread.set_inactive()

            ret = old_f(self, *args, **kwds)

            # start timeout after method reaches end
            if timeout is not None:
                # logger.debug("restarting timeout: %ds", timeout)
                cb_thread.retrigger(self, timeout, timeout_cb)

            return ret

        if state_type == STATE_TYPE_BEGIN:
            new_f._state_method_begin = True
        return new_f

    return gen


class AutomateMeta(type):
    def __new__(cls, clsname, clsbases, clsdict):
        t = type.__new__(cls, clsname, clsbases, clsdict)
        for key, val in clsdict.items():
            state_method = getattr(val, "_state_method_begin", None)

            if state_method is not None:
                t._state = key
                break
        return t


class StateMachine(object, metaclass=AutomateMeta):
    """
    This state machine allows to react on network stimulus (incoming packets)
    and imitate/build protocols.

    State_1 -> event: decide next state -> State_2 ...
    """

    def __init__(self, receive_cb):
        self._states = set()
        self._actions = set()
        self._receive_cb = receive_cb
        self._is_running = True
        self._old_f = None
        self._state = getattr(self, self._state, None)

        if self._state is None:
            logger.exception("no initial state defined!")

        self._receive_thread = threading.Thread(
            target=StateMachine.receive_cycler,
            args=[self]
        )
        self._receive_thread.start()

    @staticmethod
    def receive_cycler(obj):
        while obj._is_running:
            pkt = obj._receive_cb()

            try:
                obj._state(pkt)
            except:
                logger.warning(
                    "could not execute callback: %r",
                    obj._state
                )
        # ex.printstacktrace()

    # logger.debug("receive cycler finished")

    def stop(self):
        self._is_running = False
        # _receive_cb() (eg sockets) needs to be stopped first or this will likely hang
        self._receive_thread.join()

        try:
            _cb_thread = _cb_threads.get(self.__class__, None)
            _cb_thread.stop()
        except AttributeError:
            pass
        # logger.debug("no cb thread found")
        except Exception as ex:
            ex.printstacktrace()



"""Spanning Tree Protocol."""

from packetracer import packetracer


class STP(packetracer.Packet):
    __hdr__ = (
        ("proto_id", "H", 0),
        ("v", "B", 0),
        ("type", "B", 0),
        ("flags", "B", 0),
        ("root_id", "8s", b""),
        ("root_path", "I", 0),
        ("bridge_id", "8s", b""),
        ("port_id", "H", 0),
        ("age", "H", 0),
        ("max_age", "H", 0),
        ("hello", "H", 0),
        ("fd", "H", 0)
    )



import logging

logger = logging.getLogger("packetracer")


class LazyDict(dict):
    def __init__(self, cb_createentries):
        super().__init__()
        self._cb_createentries = cb_createentries

    def _lazy_init(self):
        if self._cb_createentries is None:
            return

        new_entries = self._cb_createentries()
        # logger.debug("Updating dict: %s" % new_entries)
        self.update(new_entries)
        self._cb_createentries = None

    # Python predefined overwritten methods
    # TODO: not all functions overwritten yet
    def __contains__(self, key):
        self._lazy_init()
        return super().__contains__(key)

    def __getitem__(self, key):
        self._lazy_init()
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        self._lazy_init()
        super().__setitem__(key, value)

    def __delitem__(self, key):
        self._lazy_init()
        pass

    def __str__(self):
        self._lazy_init()
        return super().__str__()

    def __iter__(self):
        self._lazy_init()
        super().__iter__()

    def __len__(self):
        self._lazy_init()
        return super().__len__()

    def items(self):
        self._lazy_init()
        return super().items()



# -*- coding: utf-8 -*-
#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(name="packetracer",
      version="1.2",
      author="Arash Hatami",
      author_email="hatamiarash7@gmail.com",
      url="https://github.com/hatamiarash7/PacketTracer",
      description="Packet Tracer: The fast and simple packet creating and parsing module",
      license="MIT",
      packages=[
          "packetracer",
          "packetracer.layer12",
          "packetracer.layer3",
          "packetracer.layer4",
          "packetracer.layer567"
      ],
      package_data={"packetracer": ["oui_stripped.txt"]},
      classifiers=[
          "Development Status :: 6 - Mature",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: MIT License",
          "Natural Language :: English",
          "Programming Language :: Python :: 3.3",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: Implementation :: CPython",
          "Programming Language :: Python :: Implementation :: PyPy"
      ],
      install_requires=[
		  "netifaces",
      ],
      python_requires=">=3.3.*"
      )



"""
Border Gateway Protocol.
"""
import logging

from packetracer import packetracer, triggerlist
from packetracer.structcbs import unpack_H

logger = logging.getLogger("packetracer")

# Border Gateway Protocol 4 - RFC 4271
# Communities Attribute - RFC 1997
# Capabilities - RFC 3392
# Route Refresh - RFC 2918
# Route Reflection - RFC 4456
# Confederations - RFC 3065
# Cease Subcodes - RFC 4486
# NOPEER Community - RFC 3765
# Multiprotocol Extensions - 2858

# Message Types
OPEN = 1
UPDATE = 2
NOTIFICATION = 3
KEEPALIVE = 4
ROUTE_REFRESH = 5

# Attribute Types
ORIGIN = 1
AS_PATH = 2
NEXT_HOP = 3
MULTI_EXIT_DISC = 4
LOCAL_PREF = 5
ATOMIC_AGGREGATE = 6
AGGREGATOR = 7
COMMUNITIES = 8
ORIGINATOR_ID = 9
CLUSTER_LIST = 10
MP_REACH_NLRI = 14
MP_UNREACH_NLRI = 15

# Origin Types
ORIGIN_IGP = 0
ORIGIN_EGP = 1
INCOMPLETE = 2

# AS Path Types
AS_SET = 1
AS_SEQUENCE = 2
AS_CONFED_SEQUENCE = 3
AS_CONFED_SET = 4

# Reserved Communities Types
NO_EXPORT = 0xFFFFFF01
NO_ADVERTISE = 0xFFFFFF02
NO_EXPORT_SUBCONFED = 0xFFFFFF03
NO_PEER = 0xFFFFFF04

# Common AFI types
AFI_IPV4 = 1
AFI_IPV6 = 2

# Multiprotocol SAFI types
SAFI_UNICAST = 1
SAFI_MULTICAST = 2
SAFI_UNICAST_MULTICAST = 3

# OPEN Message Optional Parameters
AUTHENTICATION = 1
CAPABILITY = 2

# Capability Types
CAP_MULTIPROTOCOL = 1
CAP_ROUTE_REFRESH = 2

# NOTIFICATION Error Codes
MESSAGE_HEADER_ERROR = 1
OPEN_MESSAGE_ERROR = 2
UPDATE_MESSAGE_ERROR = 3
HOLD_TIMER_EXPIRED = 4
FSM_ERROR = 5
CEASE = 6

# Message Header Error Subcodes
CONNECTION_NOT_SYNCHRONIZED = 1
BAD_MESSAGE_LENGTH = 2
BAD_MESSAGE_TYPE = 3

# OPEN Message Error Subcodes
UNSUPPORTED_VERSION_NUMBER = 1
BAD_PEER_AS = 2
BAD_BGP_IDENTIFIER = 3
UNSUPPORTED_OPTIONAL_PARAMETER = 4
AUTHENTICATION_FAILURE = 5
UNACCEPTABLE_HOLD_TIME = 6
UNSUPPORTED_CAPABILITY = 7

# UPDATE Message Error Subcodes
MALFORMED_ATTRIBUTE_LIST = 1
UNRECOGNIZED_ATTRIBUTE = 2
MISSING_ATTRIBUTE = 3
ATTRIBUTE_FLAGS_ERROR = 4
ATTRIBUTE_LENGTH_ERROR = 5
INVALID_ORIGIN_ATTRIBUTE = 6
AS_ROUTING_LOOP = 7
INVALID_NEXT_HOP_ATTRIBUTE = 8
OPTIONAL_ATTRIBUTE_ERROR = 9
INVALID_NETWORK_FIELD = 10
MALFORMED_AS_PATH = 11

# Cease Error Subcodes
MAX_NUMBER_OF_PREFIXES_REACHED = 1
ADMINISTRATIVE_SHUTDOWN = 2
PEER_DECONFIGURED = 3
ADMINISTRATIVE_RESET = 4
CONNECTION_REJECTED = 5
OTHER_CONFIGURATION_CHANGE = 6
CONNECTION_COLLISION_RESOLUTION = 7
OUT_OF_RESOURCES = 8


class BGP(packetracer.Packet):
    __hdr__ = (
        ("marker", "16s", b"\xff" * 16),
        ("len", "H", 0),
        ("type", "B", OPEN)
    )

    def _dissect(self, buf):
        htype = buf[18]
        self._init_handler(htype, buf[19:])
        return 19

    class Open(packetracer.Packet):
        __hdr__ = (
            ("v", "B", 4),
            ("asn", "H", 0),
            ("holdtime", "H", 0),
            ("identifier", "I", 0),
            ("param_len", "B", 0),
            ("params", None, triggerlist.TriggerList)
        )

        def _dissect(self, buf):
            # logger.debug("parsing Parameter")
            pcount = buf[9]
            off = 10

            while pcount > 0:
                plen = buf[off + 2]
                param = self.Parameter(buf[off:off + plen])
                self.params.append(param)
                pcount -= 1
                off += plen
            return off

        class Parameter(packetracer.Packet):
            __hdr__ = (
                ("type", "B", 0),
                ("len", "B", 0)
            )

    class Update(packetracer.Packet):
        __hdr__ = (
            ("withdrawnlen", "H", 0),
            ("pathlen", "H", 0),
            ("wroutes", None, triggerlist.TriggerList),
            ("pathattrs", None, triggerlist.TriggerList),
            ("anncroutes", None, triggerlist.TriggerList),
        )

        def _dissect(self, buf):
            # Withdrawn Routes
            # logger.debug("Dissecting BGP update")
            off = 4
            off_end = off + unpack_H(buf[:2])[0]

            while off < off_end:
                rlen = 3 + 0
                route = Route(buf[off:])
                self.wroutes.append(route)
                off += rlen

            # Path attributes
            off_end = off + unpack_H(buf[2:4])[0]
            # logger.debug("Unpacking attributes")

            while off < off_end:
                alen = 3 + buf[off + 2]
                # logger.debug("bytes for attribute: %r" % buf[off: off + alen])
                attr = BGP.Update.Attribute(buf[off: off + alen])
                self.pathattrs.append(attr)
                off += alen

            # Announced routes
            off_end = len(buf)
            # logger.debug("Unpacking routes")

            while off + 3 <= off_end:
                rlen = 3 + 0
                # logger.debug("bytes for route: %r" % buf[off:off + rlen])
                route = Route(buf[off: off + rlen])
                self.anncroutes.append(route)
                off += rlen
            # logger.debug("Finished, off=%d off_end=%d" % (off, off_end))
            return off

        class Attribute(packetracer.Packet):
            __hdr__ = (
                ("flags", "B", 0),
                ("type", "B", 0),
                ("len", "B", 0)
            )

            def __get_o(self):
                return (self.flags >> 7) & 0x1

            def __set_o(self, o):
                self.flags = (self.flags & ~0x80) | ((o & 0x1) << 7)

            optional = property(__get_o, __set_o)

            def __get_t(self):
                return (self.flags >> 6) & 0x1

            def __set_t(self, t):
                self.flags = (self.flags & ~0x40) | ((t & 0x1) << 6)

            transitive = property(__get_t, __set_t)

            def __get_p(self):
                return (self.flags >> 5) & 0x1

            def __set_p(self, p):
                self.flags = (self.flags & ~0x20) | ((p & 0x1) << 5)

            partial = property(__get_p, __set_p)

            def __get_e(self):
                return (self.flags >> 4) & 0x1

            extended_length = property(__get_e)

            def _dissect(self, buf):
                if len(buf) > 3:
                    try:
                        atype = buf[2]
                        type_instance = BGP.Update.Attribute._switch_type_attribute[atype](buf[3:])
                        self._set_bodyhandler(type_instance)
                    # any exception will lead to: body = raw bytes
                    except Exception:
                        # logger.debug("BGP > Update > Attribute failed to set handler: %s" % e)
                        pass
                return 3

            class Origin(packetracer.Packet):
                __hdr__ = (
                    ("type", "B", ORIGIN_IGP),
                )

            class ASPath(packetracer.Packet):
                __hdr__ = (
                    ("segments", None, triggerlist.TriggerList),
                )

                def _dissect(self, buf):
                    off = 1
                    buflen = len(buf)

                    while off < buflen:
                        seglen = buf[off + 2]
                        seg = self.ASPathSegment(buf[off + 1:seglen])
                        self.segments.append(seg)
                        off += seglen
                    return off

                class ASPathSegment(packetracer.Packet):
                    __hdr__ = (
                        ("type", "B", 0),
                        ("len", "B", 0)
                    )

            class NextHop(packetracer.Packet):
                __hdr__ = (
                    ("ip", "I", 0),
                )

            class MultiExitDisc(packetracer.Packet):
                __hdr__ = (
                    ("value", "I", 0),
                )

            class LocalPref(packetracer.Packet):
                __hdr__ = (
                    ("value", "I", 0),
                )

            class AtomicAggregate(packetracer.Packet):
                pass

            class Aggregator(packetracer.Packet):
                __hdr__ = (
                    ("asn", "H", 0),
                    ("ip", "I", 0)
                )

            class OriginatorID(packetracer.Packet):
                __hdr__ = (
                    ("value", "I", 0),
                )

            class ClusterList(packetracer.Packet):
                pass

            class MPReachNLRI(packetracer.Packet):
                __hdr__ = (
                    ("afi", "H", AFI_IPV4),
                    ("safi", "B", SAFI_UNICAST),
                )

            class MPUnreachNLRI(packetracer.Packet):
                __hdr__ = (
                    ("afi", "H", AFI_IPV4),
                    ("safi", "B", SAFI_UNICAST),
                )

            class Communitie(packetracer.Packet):
                pass

            _switch_type_attribute = {
                ORIGIN: Origin,
                AS_PATH: ASPath,
                NEXT_HOP: NextHop,
                MULTI_EXIT_DISC: MultiExitDisc,
                LOCAL_PREF: LocalPref,
                ATOMIC_AGGREGATE: AtomicAggregate,
                AGGREGATOR: Aggregator,
                COMMUNITIES: Communitie,
                ORIGINATOR_ID: OriginatorID,
                CLUSTER_LIST: ClusterList,
                MP_REACH_NLRI: MPReachNLRI,
                MP_UNREACH_NLRI: MPUnreachNLRI
            }

    class Notification(packetracer.Packet):
        __hdr__ = (
            ("code", "B", 0),
            ("subcode", "B", 0),
        )

    class Keepalive(packetracer.Packet):
        pass

    class RouteRefresh(packetracer.Packet):
        __hdr__ = (
            ("afi", "H", AFI_IPV4),
            ("rsvd", "B", 0),
            ("safi", "B", SAFI_UNICAST)
        )

    __handler__ = {
        OPEN: Open,
        UPDATE: Update,
        NOTIFICATION: Notification,
        KEEPALIVE: Keepalive,
        ROUTE_REFRESH: RouteRefresh
    }


class Route(packetracer.Packet):
    __hdr__ = (
        ("len", "B", 0),
    )



"""
Wraper for TUN/TAP interfaces.
https://www.kernel.org/doc/Documentation/networking/tuntap.txt

packets written to /dev/net/tun look like "outer network -> tunX" (coming for another network)
and get handled by the kernel state machine.

> Prerequisites
mkdir /dev/net
# Create the character device /dev/net/XXX and let it point to major number 10, minor number 200.
# The devnode has to be used in "TuntapInterface -> devnode": When "/dev/net/xxx" is used this
# will create the interface xxx0
mknod /dev/net/tunA c 10 200
mknod /dev/net/tunB c 10 200

mknod /dev/net/tapA c 10 200
mknod /dev/net/tapB c 10 200

> Test tunnel
ping4 -c 1  -I tunA0 192.168.12.35; ping4  -c 1 -I tunB0 192.168.12.34
ping4 -c 1  -I tapA0 192.168.12.35; ping4  -c 1 -I tapB0 192.168.12.34

> Notes
Routing (fd -> tun/tap -> eth -> internet -> eth -> tun/tap -> fd):
	tap works (mac_src=anything, mac_dst=tap, ip_src/dst: like tun)
	tun works (ip_src=tun, ip_dst=server)
		Advantage over tap: no icmp-messages on returning packets.

> Useful commands
ip tuntap add dev tun0 mode tun user mike group users
ip addr add 192.168.3.1/24 dev tun0
ip rule list; ip link show
ip tuntap del dev tun0 mode tun; ip tuntap del dev tun1 mode tun;
# If "ip tuntap del dev tunA0 mode tun" does not work
ip link delete tunA0
"""

import logging
import os
import pathlib
import struct
import subprocess
import threading
from os import read as os_read
from os import write as os_write

from fcntl import ioctl

from packetracer import utils
from packetracer.layer12 import ethernet
from packetracer.layer3 import ip

logger = logging.getLogger("packetracer")

# Some constants used to ioctl the device file
TUNSETIFF = 0x400454CA
TUNSETOWNER = TUNSETIFF + 2
SG_SET_TIMEOUT = 0x2201
IFF_TUN = 0x0001
IFF_TAP = 0x0002
#  The kernel adds a 4-byte preamble to the frame, avoid this
# TODO: set lowest layer based on meta info
"""
3.2 Frame format:
If flag IFF_NO_PI is not set each frame format is:
	Flags [2 bytes]
	Proto [2 bytes]
	Raw protocol(IP, IPv6, etc) frame.
"""
IFF_NO_PI = 0x1000

TYPE_TUN = 0
TYPE_TAP = 1

TYPE_STR_DCT = {TYPE_TUN: "tun", TYPE_TAP: "tap"}


def exec_syscmd(cmd):
    output = subprocess.getoutput(cmd)
    logger.info(output)


class TuntapInterface(object):
    def __init__(self,
                 iface_name,
                 devnode="/dev/net/tunA",
                 ifacetype=TYPE_TUN,
                 ip_src="12.34.56.1",
                 ip_dst="12.34.56.2",
                 is_local_tunnel=False):
        """
        iface_name -- name of the local interface to be used. See devnode.
        devnode -- Path to the devnoce used for this interface. /dev/net/tunA will result in iface_name tunA0, tunA1 etc.
        ifacetype -- TYPE_TUN or TYPE_TAP
        ip_src -- Local IP address of the interface iface_name (/32 address will be used)
        ip_dst -- Remote connection point of the local interface iface_name (/32 address will be used)
        is_local_tunnel -- True is endpoint is also local, otherwise False
        """

        self._closed = False
        self._iface_name = iface_name
        self._devnode = devnode
        self._is_newly_created = False
        self._ifacetype = ifacetype
        self._is_local_tunnel = is_local_tunnel

        TuntapInterface.create_devnode(devnode)

        # Open TUN or TAP device file
        self._iface_fd = open(devnode, "r+b", buffering=0)
        tuntap_opt = IFF_TUN if ifacetype == TYPE_TUN else IFF_TAP
        self._ifr = struct.pack("16sH", iface_name.encode("UTF-8"), tuntap_opt | IFF_NO_PI)
        # Connect interface name with file descriptor. Creates the actual network interface.
        ioctl(self._iface_fd, TUNSETIFF, self._ifr)
        # ioctl(self._iface_fd, SG_SET_TIMEOUT, 1000)
        self._fileno_iface_fd = self._iface_fd.fileno()
        # Optionally, we want it be accessed by the normal user.
        # ioctl(self._iface_fd, TUNSETOWNER, 1000)

        if ip_src is not None and ip_dst is not None:
            TuntapInterface.configure_interface(iface_name, ip_src, ip_dst, is_local_tunnel=is_local_tunnel)
        utils.set_interface_state(iface_name, state_active=True)

    is_newly_created = property(lambda self: self._is_newly_created)

    @staticmethod
    def create_devnode(devnode):
        """
        Create the given devnode if not already present.
        devnode -- Name of the devnode which will be created: "/dev/net/devnode"
        """
        if pathlib.Path(devnode).exists():
            # logger.debug("devnode %s already exists" % devnode)
            return

        # logger.debug("Creating devnode: %s" % devnode)
        exec_syscmd("mkdir /dev/net")
        exec_syscmd("mknod %s c 10 200" % devnode)

    @staticmethod
    def configure_interface(iface_name, ip_src, ip_dst, is_local_tunnel=False):
        """
        is_local_tunnel -- Adjust routing rules so that this interface can be used locally (eg tun0 <-> tun1
            where tun0 and tun1 are both local interfaces)
        """
        # Wait for interface to be created
        # time.sleep(1)
        # output = exec_syscmd("ifconfig %s %s/24" % (iface_name, ip_src))
        # output = exec_syscmd("ifconfig %s %s/24 pointopoint %s" % (iface_name, ip_src, ip_dst))
        # output = exec_syscmd("ifconfig %s %s pointopoint %s" % (iface_name, ip_src, ip_dst))
        exec_syscmd("ifconfig %s %s/24" % (iface_name, ip_src))

        if is_local_tunnel:
            # Packet with target ip_dst goes through "lo" if ip_dst is on the same host.
            # Avoid this by removing local rules
            exec_syscmd("ip route del %s table local" % ip_src)
            # pointopoint creates implicit rule in "main"
            # Problem if src/dst tun are on the same host: packets pop out of tun1 (target), but the kernel
            # does not recognize them as being addressed to the local host. (we removed the rule above)
            # Solution: distinct routing decisions and configure routing in such a way that the local
            # type routes are only "seen" by the input routing decision
            tid = 13
            exec_syscmd("ip route add local %s dev %s table %d" % (ip_src, iface_name, tid))
            # make sure previous rules have been removed
            # iif NAME = select the incoming device to match
            # http://man7.org/linux/man-pages/man8/ip-rule.8.html
            exec_syscmd("ip rule del iif %s lookup %d" % (iface_name, tid))
            exec_syscmd("ip rule add iif %s lookup %d" % (iface_name, tid))

    def read(self):
        """Read an IP packet been sent to this TUN device."""
        try:
            return os_read(self._fileno_iface_fd, 1024 * 4)
        except TypeError:
            # read after closing
            return None

    def write(self, bts):
        """Write an IP packet to this TUN device."""
        try:
            os_write(self._fileno_iface_fd, bts)
        except TypeError:
            # write after closing
            pass

    def close(self):
        if self._closed:
            return
        self._closed = True

        try:
            # self._iface_fd.close()
            os.close(self._fileno_iface_fd)
            self._fileno_iface_fd = None
        except Exception as ex:
            print(ex)

        if self._is_local_tunnel:
            exec_syscmd("ip rule del iif %s lookup %d" % (self._iface_name, 13))


class LocalTunnel(object):
    """
    Local Back-to-back tunnel based on tun interfaces: local <-> ip:tun1:dev <-> dev:tun2:ip <-> local
    """

    def __init__(self, ip_iface_A="192.168.2.1", ip_iface_B="192.168.3.1"):
        self._ifacetype = TYPE_TAP
        islocaltunnel = True
        ifacetype_str = TYPE_STR_DCT[self._ifacetype]
        self._state_active = False

        iface_name_A = ifacetype_str + "A0"
        self._dev_A = TuntapInterface(
            iface_name=iface_name_A,
            devnode="/dev/net/" + ifacetype_str + "A",
            ifacetype=self._ifacetype,
            ip_src=ip_iface_A,
            is_local_tunnel=islocaltunnel
        )
        iface_name_B = ifacetype_str + "B0"
        self._dev_B = TuntapInterface(
            iface_name=iface_name_B,
            devnode="/dev/net/" + ifacetype_str + "B",
            ifacetype=self._ifacetype,
            ip_src=ip_iface_B,
            is_local_tunnel=islocaltunnel
        )

        utils.flush_arp_cache()
        # mac_A = utils.get_mac_for_iface(iface_name_A)
        # mac_B = utils.get_mac_for_iface(iface_name_B)
        # utils.add_arp_entry(ip_iface_A, mac_A, iface_name_B)
        # utils.add_arp_entry(ip_iface_B, mac_B, iface_name_A)

        self._rs_thread_A = None
        self._rs_thread_B = None

    def _start_cycler_threads(self):
        self._rs_thread_A = threading.Thread(target=LocalTunnel.read_write_cycler,
                                             args=[self, self._dev_A, self._dev_B, "1to2"])
        self._rs_thread_B = threading.Thread(target=LocalTunnel.read_write_cycler,
                                             args=[self, self._dev_B, self._dev_A, "2to1"])
        self._rs_thread_A.start()
        self._rs_thread_B.start()

    @staticmethod
    def read_write_cycler(obj, iface_in, iface_out, name):
        while obj._state_active:
            try:
                bts = iface_in.read()
                try:
                    ip.IP(bts) if obj._ifacetype == TYPE_TUN else ethernet.Ethernet(bts)
                    # logger.debug("Sending in cycler %s (%s -> %s):\n%s\n%s" %
                    #	(name, iface_in._iface_name, iface_out._iface_name, bts, pkt))
                    iface_out.write(bts)
                except:
                    pass
            except ValueError as ex:
                logger.exception(ex)
                break
            except OSError as ex:
                logger.exception(ex)
                break
            except Exception as ex:
                logger.exception(ex)
                break

    def set_state(self, state_active):
        if self._state_active is None:
            return

        if state_active == self._state_active:
            return

        self._state_active = state_active

        if state_active:
            self._start_cycler_threads()
        else:
            self._dev_A.close()
            self._dev_B.close()

            for th in [self._rs_thread_A, self._rs_thread_B]:
                try:
                    th.join()
                except:
                    pass

            self._state_active = None



"""
Internet Protocol version 6..for whoever needs it (:

RFC 2460
"""
import logging

from packetracer import packetracer, triggerlist
# handler
from packetracer.layer3 import esp, icmp6, igmp, ipx, ospf, pim
from packetracer.layer3.ip_shared import IP_PROTO_HOPOPTS, IP_PROTO_ROUTING, IP_PROTO_FRAGMENT, \
    IP_PROTO_AH, IP_PROTO_ESP, IP_PROTO_DSTOPTS, IP_PROTO_ICMP6, IP_PROTO_IGMP, IP_PROTO_TCP, \
    IP_PROTO_UDP, IP_PROTO_PIM, IP_PROTO_IPXIP, IP_PROTO_SCTP, IP_PROTO_OSPF
from packetracer.layer4 import tcp, udp, sctp
from packetracer.packetracer_meta import FIELD_FLAG_AUTOUPDATE

logger = logging.getLogger("packetracer")

ext_hdrs = {
    IP_PROTO_HOPOPTS,
    IP_PROTO_ROUTING,
    IP_PROTO_FRAGMENT,
    IP_PROTO_AH,
    IP_PROTO_ESP,
    IP_PROTO_DSTOPTS,
    # TODO: to be implemented
    # IP_PROTO_MOBILITY
    # IP_PROTO_NONEXT
}


class IP6(packetracer.Packet):
    __hdr__ = (
        ("v_fc_flow", "I", 0x60000000),
        ("dlen", "H", 0, FIELD_FLAG_AUTOUPDATE),  # length of extension header (opts header) + body
        # body handler type OR type of first extension hedader (opts header)
        ("nxt", "B", 0),
        ("hlim", "B", 0),  # hop limit
        ("src", "16s", b"\x00" * 16),
        ("dst", "16s", b"\x00" * 16),
        ("opts", None, triggerlist.TriggerList)
    )

    def __get_v(self):
        return self.v_fc_flow >> 28

    def __set_v(self, v):
        self.v_fc_flow = (self.v_fc_flow & ~0xF0000000) | (v << 28)

    v = property(__get_v, __set_v)

    def __get_fc(self):
        return (self.v_fc_flow >> 20) & 0xFF

    def __set_fc(self, v):
        self.v_fc_flow = (self.v_fc_flow & ~0xFF00000) | (v << 20)

    fc = property(__get_fc, __set_fc)

    def __get_flow(self):
        return self.v_fc_flow & 0xFFFFF

    def __set_flow(self, v):
        self.v_fc_flow = (self.v_fc_flow & ~0xFFFFF) | (v & 0xFFFFF)

    flow = property(__get_flow, __set_flow)

    # Convenient access for: src[_s], dst[_s]
    src_s = packetracer.get_property_ip6("src")
    dst_s = packetracer.get_property_ip6("dst")
    nxt_t = packetracer.get_property_translator("nxt", "IP_PROTO_")

    __handler__ = {
        IP_PROTO_ICMP6: icmp6.ICMP6,
        IP_PROTO_IGMP: igmp.IGMP,
        IP_PROTO_TCP: tcp.TCP,
        IP_PROTO_UDP: udp.UDP,
        IP_PROTO_ESP: esp.ESP,
        IP_PROTO_PIM: pim.PIM,
        IP_PROTO_IPXIP: ipx.IPX,
        IP_PROTO_SCTP: sctp.SCTP,
        IP_PROTO_OSPF: ospf.OSPF
    }

    update_dependants = {tcp.TCP, udp.UDP}

    def _dissect(self, buf):
        type_nxt = buf[6]
        off = 40
        opts = []

        # logger.debug("type: %d buflen: %d" % (type_nxt, len(buf)))
        # logger.debug("parsing opts from bytes (dst: %s): (len: %d) %s" %
        #	 (buf[24:40], self.hdr_len, buf[off:]))
        # parse options until type is an upper layer one
        while type_nxt in ext_hdrs:
            length = 8 + buf[off + 1] * 8
            # logger.debug("next type is: %s, len: %d, %r" % (type_nxt, length, buf[off:off + length]))
            opt = ext_hdrs_cls[type_nxt](buf[off:off + length])
            opts.append(opt)
            type_nxt = buf[off]
            off += length

        # TODO: lazy dissect possible?
        self.opts.extend(opts)
        # IPv6 and IPv4 share same handler
        # There are some cases where padding can not be identified on ethernet -> do it here (eg VSS shit trailer)
        self._init_handler(type_nxt, buf[off:])
        return off

    def _update_fields(self):
        if self.dlen_au_active:
            self.dlen = len(self.opts.bin()) + len(self.body_bytes)
        # Set type value in nxt OR in last opts element (if present)
        # Updating is a bit more complicated so we can't use FIELD_FLAG_IS_TYPEFIELD
        # idval is None if body handler is None
        # logger.debug("handler %r -> %r", self.__class__, self.higher_layer.__class__)
        idval = packetracer.Packet.get_id_for_handlerclass(self.__class__, self.higher_layer.__class__)
        # logger.debug("nxt will be %r", idval)

        if idval is not None:
            if len(self.opts) == 0:
                self.nxt = idval
            else:
                # problem if opts[-1] is immutable
                try:
                    self.opts[-1].nxt = idval
                except:
                    pass

    def direction(self, other):
        # logger.debug("checking direction: %s<->%s" % (self, next))
        if self.src == other.src and self.dst == other.dst:
            # consider packet to itself: can be DIR_REV
            return packetracer.Packet.DIR_SAME | packetracer.Packet.DIR_REV
        if self.src == other.dst and self.dst == other.src:
            return packetracer.Packet.DIR_REV
        return packetracer.Packet.DIR_UNKNOWN

    def reverse_address(self):
        self.src, self.dst = self.dst, self.src


#
# Basic shared option classes
#
class IP6OptsHeader(packetracer.Packet):
    __hdr__ = (
        ("nxt", "B", 0),  # next extension header protocol
        ("len", "B", 0),  # option data length in 8 octect units (ignoring first 8 octets) so, len 0 == 64bit header
        ("opts", None, triggerlist.TriggerList)
    )

    def _dissect(self, buf):
        length = 8 + buf[1] * 8
        options = []
        off = 2

        while off < length:
            opt_type = buf[off]
            # logger.debug("IP6OptsHeader: type: %d" % opt_type)

            # http://tools.ietf.org/html/rfc2460#section-4.2
            # PAD1 option: no length or data field
            if opt_type == 0:
                opt = IP6OptionPad(type=opt_type)
                # logger.debug("next ip6 bytes 1: %r" % (buf[off:off + 2]))
                off += 1
            else:
                opt_len = buf[off + 1]
                opt = IP6Option(type=opt_type, len=opt_len, body_bytes=buf[off + 2: off + 2 + opt_len])
                # logger.debug("next ip6 bytes 2: %r" % (buf[off + 2: off + 2 + opt_len]))
                off += 2 + opt_len
            options.append(opt)

        self.opts.extend(options)
        return off


class IP6Option(packetracer.Packet):
    __hdr__ = (
        ("type", "B", 0),
        ("len", "B", 0)
    )


class IP6OptionPad(packetracer.Packet):
    __hdr__ = (
        ("type", "B", 0),
    )


class IP6HopOptsHeader(IP6OptsHeader):
    def _dissect(self, buf):
        # logger.debug("IP6HopOptsHeader parsing")
        return IP6OptsHeader._dissect(self, buf)


class IP6RoutingHeader(packetracer.Packet):
    __hdr__ = (
        ("nxt", "B", 0),  # next extension header protocol
        ("len", "B", 0),  # extension data length in 8 octect units (ignoring first 8 octets) (<= 46 for type 0)
        ("type", "B", 0),  # routing type (currently, only 0 is used)
        ("segs_left", "B", 0),  # remaining segments in route, until destination (<= 23)
        ("rsvd_sl_bits", "I", 0),  # reserved (1 byte), strict/loose bitmap for addresses
        ("addresses", None, triggerlist.TriggerList)
    )

    def __get_sl_bits(self):
        return self.rsvd_sl_bits & 0xFFFFFF

    def __set_sl_bits(self, v):
        self.rsvd_sl_bits = (self.rsvd_sl_bits & ~0xFFFFF) | (v & 0xFFFFF)

    sl_bits = property(__get_sl_bits, __set_sl_bits)

    def _dissect(self, buf):
        hdr_size = 8
        addr_size = 16
        addresses = []
        num_addresses = self.buf[1] / 2

        buf = buf[hdr_size:hdr_size + num_addresses * addr_size]

        # logger.debug("IP6RoutingHeader: parsing addresses")
        for i in range(num_addresses):
            addresses.append(buf[i * addr_size: i * addr_size + addr_size])

        self.addresses.extend(addresses)
        return len(num_addresses) * addr_size + addr_size
    # setattr(self, "addresses", addresses)
    # setattr(self, "length", self.len * 8 + 8)


class IP6FragmentHeader(packetracer.Packet):
    __hdr__ = (
        ("nxt", "B", 0),  # next extension header protocol
        ("resv", "B", 0),  # reserved, set to 0
        ("frag_off_resv_m", "H", 0),  # frag offset (13 bits), reserved zero (2 bits), More frags flag
        ("id", "I", 0)  # fragments id
    )

    def __get_frag_off(self):
        return self.frag_off_resv_m >> 3

    def __set_frag_off(self, v):
        self.frag_off_resv_m = (self.frag_off_resv_m & ~0xFFF8) | (v << 3)

    frag_off = property(__get_frag_off, __set_frag_off)

    def __get_m_flag(self):
        return self.frag_off_resv_m & 1

    def __set_m_flag(self, v):
        self.frag_off_resv_m = (self.frag_off_resv_m & ~0xFFFE) | v

    m_flag = property(__get_m_flag, __set_m_flag)


class IP6AHHeader(packetracer.Packet):
    __hdr__ = (
        ("nxt", "B", 0),  # next extension header protocol
        ("len", "B", 0),  # length of header in 4 octet units (ignoring first 2 units)
        ("resv", "H", 0),  # reserved, 2 bytes of 0
        ("spi", "I", 0),  # SPI security parameter index
        ("seq", "I", 0)  # sequence no.
    )


class IP6ESPHeader(packetracer.Packet):
    def _dissect(self, buf):
        raise NotImplementedError("ESP extension headers are not supported.")


class IP6DstOptsHeader(IP6OptsHeader):
    def _dissect(self, buf):
        # logger.debug("IP6DstOptsHeader parsing")
        IP6OptsHeader._dissect(self, buf)


ext_hdrs_cls = {
    IP_PROTO_HOPOPTS: IP6HopOptsHeader,
    IP_PROTO_ROUTING: IP6RoutingHeader,
    IP_PROTO_FRAGMENT: IP6FragmentHeader,
    IP_PROTO_ESP: IP6ESPHeader,
    IP_PROTO_AH: IP6AHHeader,
    IP_PROTO_DSTOPTS: IP6DstOptsHeader
    # IP_PROTO_MOBILITY:
    # IP_PROTO_NONEXT:
}
